# The default script execution is set to Form Based Interactive option $AMIAutomation = $false, the same script is executed in the Ami Automation framework set to $true
Param (
    $AMIAutomation = $false,
    $productCode = $null
)

$ScriptVersion = "2.13.24"
# placeholder for index of the failed functions $failedIndexList. The -fix parameter will call only functions that were marked/indexed as failed
$global:failedIndexList = @()
$global:PASSED = "PASSED"
$global:WARNING = "WARNING"
$global:FAILED = "FAILED"
$global:ImageInfo = @{ }
$global:FreeSpaceMinimumGB = 10
$global:FreeSpaceMinimumGBOffice = 20
$global:MinRearmCount = 1
$global:MaxDriveSize = 80
$global:MinPSVersion = 4
$global:SuccessLabel = "[SUCCESS]"
$global:FailureLabel = "[FAIL]"
$global:FailureText = "Unable to resolve the issue, please click on the info button for more detailed information"
$global:SupportedOS = @("10.0.10586", "10.0.14393", "10.0.15063", "10.0.16299", "10.0.17134", "10.0.17763", "10.0.18362", "10.0.18363", "10.0.19041","10.0.19042","10.0.19043","10.0.19044","10.0.19045","10.0.22000","10.0.22621","10.0.22631")
$officeLink = 'https://docs.aws.amazon.com/workspaces/latest/adminguide/byol-windows-images.html'
$officeMessage = 'The BYOL Checker has detected that Microsoft Office is installed on this virtual machine (VM). During the BYOL image ingestion process, you can choose to subscribe to Microsoft Office Professional 2016 or 2019 through AWS. If you plan to subscribe to Office through AWS, you must uninstall Office from this VM, and then run BYOL Checker again.'
$officeTitle = 'Microsoft Office detected'
$officeRegistryPattern = "0FF1CE"
$officeFreeSpacemessage='The BYOL Checker has detected that there is less than 20 GB of free disk space on your virtual machine (VM). During the BYOL image ingestion process, you can choose to subscribe to Microsoft Office Professional 2016 or 2019 through AWS. If you plan to subscribe to Office through AWS, make sure you have at least 20 GB of free disk space on your VM, and then run BYOL Checker again.'


# Placeholder for Functions/Tests Exemptions that will be excluded from the execution
$WIN7ExemptedFunctions = @("Test-BYOL_AzureDomainJoined", "Test-BYOL_OfficeInstalled","Test-BYOL_FreeDiskSpace")
$WIN10ExemptedFunctions = @("Test-BYOL_FreeDiskSpaceW7")
$AutomationExemptFunctions = @("Test-BYOL_DriveLessThan80GB")
$warningExemptedFunctions = @("Test-BYOL_OfficeInstalled","Test-BYOL_FreeDiskSpace")

# placeholder for script variables
$CurrentWorkingDirectory = $PSScriptRoot
$fileTimestamp = get-Date -f yyyy'-'MM'-'dd'_'hhmmss
$LogFileName = "BYOLPrevalidationlog" + $fileTimeStamp
$Logfile = "$currentWorkingDirectory\$LogFileName.txt"
$Infofile = "$currentWorkingDirectory\ImageInfo.txt"
$scriptName = $($MyInvocation.MyCommand.Name)
$appName = $scriptName # is used in IMS
$global:sysprepTimeRun = 300 # run time sysprep
$global:OSCheck = (Get-CimInstance win32_operatingsystem).Caption
[PsCustomObject]$global:FunctionList = @()

# Get all functions that start with "Test-BYOL_*" . This function will dynamically create a list of functions that will be executed to validate the BYOL.
function Get-BYOLFunctions {
    $allFunctions = Get-ChildItem function:\
    $global:BYOLFunctions = $allFunctions | Where-Object Name -like "Test-BYOL_*" | Select-Object Name
    return $global:BYOLFunctions
}
function Test-BYOL_PowershellVersion {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "PowerShell version installed is 4.0 or higher"
                FailureText = "Upgrade to PowerShell version 4.0 or higher."
            }
            $global:FunctionList += $obj
        }
        true {
            $log1 = $PSVersionTable.PSVersion.Major
            Write-LogFile "Powershell version is $log1"
            if ($PSVersionTable.psversion.major -ge $global:MinPSVersion) {
                Write-LogFile " $global:SuccessLabel PowerShell Version is greater than $global:MinPSVersion "
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "OutdatedPowershellVersion"
                Write-LogFile " $global:FailureLabel PowerShell Version is less than $global:MinPSVersion"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }

        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-OfficeInstalled {
    # This function return true if Microsoft Office is detected on the machine
    $apps = Get-ItemProperty @('hklm:\software\microsoft\windows\currentversion\uninstall\*', 'hklm:\software\Wow6432Node\microsoft\windows\currentversion\uninstall\*') | Where { $_.PSChildName -match $officeRegistryPattern }
    if ($apps) {
        return $true
    }
    else {
        return $false
    }
}
function Test-BYOL_OfficeInstalled {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Microsoft Office check"
                FailureText = "$officeMessage Learn more: $officeLink"
            }
            $global:FunctionList += $obj
        }
        true {
            # Init Office configuration (Installed apps, SkipOfficePopup registry value)
            $officeDetected = Test-OfficeInstalled
            $skipOfficePopup = Get-ItemProperty -path "HKLM:\Software\Amazon\BYOLChecker\" -Name SkipOfficePopup -ErrorAction SilentlyContinue | Select-Object -ExpandProperty SkipOfficePopup

            # Display Popup during Interactive Run of BYOLChecker. The SkipOfficePopup registry key will be automatically removed when running BYOLChecker in interactive mode to force the display of the Popup.
            if ((-not $skipOfficePopup) -and $officeDetected) {
                Write-LogFile "Microsoft Office detected [interactiveMode]"
                [System.Reflection.Assembly]::LoadWithPartialName("System.Windows.Forms")
                [System.Windows.Forms.MessageBox]::Show("$officeMessage Click on the help button to learn more.", $officeTitle, [System.Windows.Forms.MessageBoxButtons]::OK, [System.Windows.Forms.MessageBoxIcon]::Warning, 'button3', 0, $officeLink)
                $skipOfficePopup = (New-ItemProperty -Path "HKLM:\Software\Amazon\BYOLChecker" -Name SkipOfficePopup -Value 1 -Force).skipOfficePopup
            }

            # Adding the value of the regitry in the logs for troubleshooting
            Write-LogFile "HKLM:\Software\Amazon\BYOLChecker\SkipOfficePopup = $skipOfficePopup"

            if ($officeDetected) {
                # If Microsoft Office is detected, we will try to determine if this causes an issue.
                switch ($interactiveMode) {
                    $true {
                        # Microsoft Office is detected during an interactive run of BYOLChecker. Returning WARNING in the tool.
                        Write-LogFile "$global:FailureLabel $officeMessage Learn more:$officeLink"
                        return $global:WARNING
                    }
                    default {
                        # Microsoft Office is detected during an automated run of BYOLChecker.
                        if ($productCode -match "_OFFICE_") {
                            # Failing the check if the product code is related to Microsoft Office
                            $ErrorCode = "OfficeInstalled"
                            Write-LogFile "$global:FailureLabel Unable to create BYOL image using $productCode because Microsoft Office is installed. $officeMessage Learn more: $officeLink"
                            return [tuple]::Create($global:FAILED, $ErrorCode)
                        }
                        else {
                            # Passing the check if the product code is not related to Microsoft Office.
                            Write-LogFile "$global:SuccessLabel"
                            return [tuple]::Create($global:PASSED)
                        }
                    }
                }
            }
            else {
                # Passing the check if Microsoft Office is not detected on the machine.
                Write-LogFile "$global:SuccessLabel"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            # No automated fix for this check.
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_PCoIP_Check {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "PCoIP Agent is not installed"
                FailureText = "PCoIP Agent is installed. Uninstall PCoIP Agent. "
            }
            $global:FunctionList += $obj
            $global:PCoIP_del_try = $false
        }
        true {
            # Tipacally all three Checks must be false to properly install new PCoIP Agent
            $global:PCoIP_uninstaller = "C:\Program Files (x86)\Teradici\PCoIP Agent\uninst.exe"
            $Check1 = Test-Path "HKLM:\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\PCoIP*"
            $Check2 = Test-Path $PCoIP_uninstaller
            $Check3 = Get-Service | Where-Object { $_.Name -eq "PCoIPAgent" }
            $ErrorCode = "PCoIPAgentInstalled"
            if (!($Check1) -and !($Check2) -and !($Check3)) {
                # All Checks are false, no PCoIP agent artifacts
                Write-LogFile "$global:SuccessLabel PCoIP agent is not present"
                return [tuple]::Create($global:PASSED)
            }
            elseif (!($global:PCoIP_del_try) -and !($Check2)) {
                # If uninst.exe is missing it will mark as Failed at first run, because we cannot uninstall the agent
                Write-LogFile " $global:FailureLabel PCoIP uninstaller failed"
                return [tuple]::Create($global:FAILED, $ErrorCode)

            }
            elseif ($Check1 -or $Check2 -or $Check3 -or !($global:PCoIP_del_try)) {
                # If any of the Checks fail, than the PCoIP was present and an uninstall can be initiated
                Write-LogFile " $global:FailureLabel Detected PCoIP Agent installed"
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
            else {
                Write-LogFile " $global:FailureLabel Unexpected PCoIP check failure"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText("Uninstalling PCoIP Agent. It can take several minutes. Please wait ...")
            $ExitCode = (Start-Process -FilePath $global:PCoIP_uninstaller -ArgumentList "/S /NoPostReboot '_?=C:\Program Files (x86)\Teradici\PCoIP Agent'" -Wait -PassThru).ExitCode
            if ($ExitCode -eq 0) {
                $global:OutputBox.AppendText("`nPCoIP Agent has been uninstalled successfullly")
                Write-LogFile "$global:SuccessLabel PCoIP Agent has been uninstalled successfullly"
            }
            else {
                $global:OutputBox.AppendText("`nFailed to uninstall PCoIP Agent with ExitCode: $ExitCode")
                Write-LogFile "$global:FailureLabel Failed to uninstall PCoIP Agent with ExitCode: $ExitCode"
                $global:PCoIP_del_try = $true
                $global:OutputBox.AppendText("$global:FailureText `nFailed to auto uninstall PCoIP Agent. Try uninstalling it manually.")
            }
        }
    }
}
function Test-BYOL_Disable_Updates {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Windows updates is disabled"
                FailureText = "Windows updates is enabled. Disable Windows updates and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            # Desired Service Name and its StartType Mode
            $global:GoodService = @(
                [PsCustomObject]@{
                    Name      = "wuauserv"
                    StartMode = "Disabled"
                }
                ,
                [PsCustomObject]@{
                    Name      = "TrustedInstaller"
                    StartMode = "Manual"
                }
            )

            $ServiceStateList = @()
            foreach ($svc in $global:GoodService) {
                $ServiceState = Get-WmiObject -ClassName Win32_Service | Where-Object Name -eq $svc.Name | Select-Object Name, StartMode
                $ServiceStateList = $ServiceStateList + $ServiceState
            }
            $global:BadService = @()
            $global:BadService = Compare-Object $global:GoodService $ServiceStatelist -Property StartMode, Name | Select-Object Name -Unique -ExpandProperty Name
            if ($global:BadService -eq $null) {
                Write-LogFile "$global:SuccessLabel Update service status has passed the check"
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "WindowsUpdatesEnabled"
                Write-LogFile "$global:FailureLabel Update service status failed. Service(s): $global:BadService did not meet the startup criteria."
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
        }
        fix {
            foreach ($service in $global:BadService) {
                try {
                    # Stopping the process, if it is running, before disabling it
                    $ServicePID = (get-wmiobject win32_service | Where-Object { $_.name -eq $service }).processID
                    if ($ServicePID -ne 0) {
                        Write-LogFile "$global:FailureLabel $service processID is $ServicePID. Stopping the service"
                        taskkill /pid $ServicePID /f /t
                    }
                    else {
                        Write-LogFile "$global:SuccessLabel $service processID is 0 (null)"
                    }
                    # Disabling the service
                    Stop-Service -Name $service -Force -ErrorAction SilentlyContinue
                    Set-Service -Name $service -StartupType ($global:GoodService | Where-Object { $_.Name -eq $service }).StartMode
                }
                catch {
                    $global:OutputBox.AppendText("$global:FailureText `nFailed to Auto disable Windows Updates. Please try to manualy disable Windows Updates.")
                }
            }
        }
    }
}
function Test-BYOL_AutoMount {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Automount is enabled"
                FailureText = "Automount is disabled. Enable Automount."
            }
            $global:FunctionList += $obj
        }
        true {
            $RegKey = "HKLM:\SYSTEM\CurrentControlSet\services\mountmgr"

            if ((Get-ItemProperty $RegKey).NoAutoMount -eq 1) {
                $ErrorCode = "AutoMountDisabled"
                Write-LogFile " $global:FailureLabel Automount is Disabled."
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
            else {
                Write-LogFile " $global:SuccessLabel Automount is Enabled"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            "automount enable" | diskpart
        }
    }
}
function Test-BYOL_Workspaces_BYOLAccountExist {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "WorkSpaces_BYOL account exists"
                FailureText = "WorkSpaces_BYOL account not found in the SAM database. Create an account with this username and add it to local administrators group."
            }
            $global:FunctionList += $obj
        }
        true {
            if (Get-WMIObject Win32_UserAccount -Filter "LocalAccount='true' and Name='WorkSpaces_BYOL'") {

                Write-LogFile " Workspaces_BYOL Exists"
                Write-LogFile "$global:SuccessLabel Workspaces_BYOL Account Exists"
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "WorkspacesBYOLAccountNotFound"
                Write-LogFile " $global:FailureLabel Workspaces_BYOL Account Does Not Exist"
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
        }
        fix {
            New-WorkSpaces_BYOL_User
        }
    }
}
function Test-BYOL_Workspaces_BYOLAccountDisabled {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "WorkSpaces_BYOL account is enabled"
                FailureText = "WorkSpaces_BYOL account is disabled. Please enable it."
            }
            $global:FunctionList += $obj
        }
        true {
            $script:workspaces_BYOLAccount = Get-WmiObject -Class Win32_UserAccount -Filter  "LocalAccount='True'" | Where-Object { $_.Name -eq "WorkSpaces_BYOL" }
            $log14 = $workspaces_BYOLAccount.Disabled
            Write-LogFile "WorkSpaces_BYOL Disabled/Enabled status - $log14 (Status will be blank if this test is run on domain controller)"
            $ImageInfo["WorkSpaces_BYOLAccount.Disabled"] = $log14
            if ($workspaces_BYOLAccount.Disabled -eq $false) {
                Write-LogFile "$global:SuccessLabel WorkSpaces_BYOL Account Enabled"
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "WorkspacesBYOLAccountDisabled"
                Write-LogFile " $global:FailureLabel WorkSpaces_BYOL Account Disabled"
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
        }
        fix {
            net user Workspaces_BYOL /active:yes
        }
    }
}
function Test-BYOL_DHCPEnabledInterface {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "DHCP is enabled on network interface"
                FailureText = "Network interface is currently using a static IP address. Change network interface to use DHCP."
            }
            $global:FunctionList += $obj
        }
        true {
            $script:networkAdapters = Get-WmiObject -Class win32_networkadapterconfiguration -filter 'ipenabled = "true"'
            $log = $networkAdapters.DHCPEnabled
            Write-LogFile "Network Adapter using DHCP logs - $log"
            $ImageInfo["NetworkAdapterUsingDHCP"] = $log
            #should not have 2 network adapters but we want to treat this test separate and consider multiple network interfaces but will succeed if true on all interfaces.
            if ($log -eq $False) {
                $ErrorCode = "DHCPDisabled"
                Write-LogFile " $global:FailureLabel DHCP disabled on network interface"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel DHCP enabled on network interface"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_FreeDiskSpaceW7 {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "More than ${global:FreeSpaceMinimumGB}GB of free space on C: drive"
                FailureText = "C: drive has less than ${global:FreeSpaceMinimumGB}GB of free space. Clean up C: drive to free up some space"
            }
            $global:FunctionList += $obj
        }
        true {
            $script:disk = Get-WmiObject Win32_LogicalDisk -Filter "DeviceID='C:'" | Select-Object FreeSpace
            $currentFreeSpace = ($($disk.freespace)/1GB)
            Write-LogFile " Amount of free space on C: drive - $currentFreeSpace GB"
            $ImageInfo["FreeSpaceOnCDrive"] = $currentFreeSpace
            if ($currentFreeSpace -lt $global:FreeSpaceMinimumGB ) {
                $ErrorCode = "DiskFreeSpace"
                Write-LogFile "$global:FailureLabel Free disk space is less than minimum required $global:FreeSpaceMinimumGB"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel Free disk space is greater than minimum required $global:FreeSpaceMinimumGB"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_FreeDiskSpace {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "More than ${global:FreeSpaceMinimumGBOffice}GB of free space on C: drive"
                FailureText = "$officeFreeSpacemessage Learn more: $officeLink"
            }
            $global:FunctionList += $obj
        }
        true {
            $script:disk = Get-WmiObject Win32_LogicalDisk -Filter "DeviceID='C:'" | Select-Object FreeSpace
            $currentFreeSpace = [math]::Round(($($disk.freespace)/1GB),2)
            Write-LogFile "Amount of free space on C: drive - $currentFreeSpace GB"
            $ImageInfo["FreeSpaceOnCDrive"] = $currentFreeSpace
            $officeDetected = Test-OfficeInstalled
            $ErrorCode = "DiskFreeSpace"

            if ($interactiveMode) {
                if ($currentFreeSpace -lt $global:FreeSpaceMinimumGBOffice ) {
                    Write-LogFile "$global:FailureLabel Free disk space is less than minimum required to support Microsoft Office by AWS $global:FreeSpaceMinimumGBOffice GB"
                    return [tuple]::Create($global:WARNING, $ErrorCode)
                }
                elseif ($currentFreeSpace -lt $global:FreeSpaceMinimumGB ) {
                    Write-LogFile "$global:FailureLabel Free disk space is less than minimum required $global:FreeSpaceMinimumGB GB"
                    return [tuple]::Create($global:FAILED, $ErrorCode)
                }
                else {
                    Write-LogFile "$global:SuccessLabel Free disk space is greater than minimum required $global:FreeSpaceMinimumGBOffice GB"
                    return [tuple]::Create($global:PASSED)
                }
            }
            else {
                switch ($officeDetected) {
                    $true {
                        if ($productCode -match "_OFFICE_") {
                            if ($currentFreeSpace -lt $global:FreeSpaceMinimumGBOffice) {
                                Write-LogFile "$global:FailureLabel Free disk space is less than minimum required $global:FreeSpaceMinimumGBOffice GB"
                                return [tuple]::Create($global:FAILED, $ErrorCode)
                            }
                            else {
                                Write-LogFile "$global:SuccessLabel Free disk space is greater than minimum required $global:FreeSpaceMinimumGB GB"
                                return [tuple]::Create($global:PASSED)
                            }
                        }
                        else {
                            if ($currentFreeSpace -lt $global:FreeSpaceMinimumGB) {
                                Write-LogFile "$global:FailureLabel Free disk space is less than minimum required $global:FreeSpaceMinimumGB GB"
                                return [tuple]::Create($global:FAILED, $ErrorCode)
                            }
                            else {
                                Write-LogFile "$global:SuccessLabel Free disk space is greater than minimum required $global:FreeSpaceMinimumGB GB"
                                return [tuple]::Create($global:PASSED)
                            }
                        }
                    }
                    default {
                        if ($currentFreeSpace -lt $global:FreeSpaceMinimumGB) {
                            Write-LogFile "$global:FailureLabel Free disk space is less than minimum required $global:FreeSpaceMinimumGB GB"
                            return [tuple]::Create($global:FAILED, $ErrorCode)
                        }
                        else {
                            Write-LogFile "$global:SuccessLabel Free disk space is greater than minimum required $global:FreeSpaceMinimumGB GB"
                            return [tuple]::Create($global:PASSED)
                        }
                    }

                }
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_LocalDrives {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Only local disks are attached"
                FailureText = "Removable or network drives attached. Remove all removable, network, and CD/ISO drives."
            }
            $global:FunctionList += $obj
        }
        true {
            $script:drives = @(Get-WmiObject -class win32_logicaldisk | Select-Object DeviceID, DriveType)
            $log9 = $drives | Out-String
            Write-LogFile " Attached Drives types:"
            Write-LogFile "$log9"
            if ($drives.Count -eq "1" -and $drives.DriveType -eq "3") {
                # Only one drive should be attached and it has to be Local Drive
                Write-LogFile "$global:SuccessLabel Drives of the correct type attached."
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "AdditionalDrivesAttached"
                Write-LogFile " $global:FailureLabel Wrong type of drives attached."
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_OSRequirements {
    # Reference for version https://en.wikipedia.org/wiki/Ver_%28command%29
    # Reference for edition numbers https://techontip.wordpress.com/tag/operatingsystemsku/
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Supported Windows operating system"
                FailureText = "Operating System is not supported. Try again with image running supported OS."
            }
            $global:FunctionList += $obj
        }
        true {
            $OSinfo = get-wmiobject -class win32_operatingsystem
            [int]$OSType = $OSinfo.ProductType # 1 is for Workstation
            [String]$OSVersion = $OSinfo.Version
            # Only Workstations and WIN7, WIN10, WIN11 can be used for BYOL
            # ProductType --> 1 for client versions of Windows, 2 for server versions of Windows operating as domain controllers, and 3 for server versions of Windows that are not operating as domain controllers.
            if ($osType -eq 1 -and ($OSVersion -in $global:SupportedOS)) {
                Write-LogFile "$global:SuccessLabel OS type $osType is Desktop type and OS Version is $OSVersion"
                $global:OOBEfile = "$PSScriptRoot\OOBE_unattend.xml"
                return [tuple]::Create($global:PASSED)
            }
           else {
                $ErrorCode = "OSNotSupported"
                Write-LogFile " $global:FailureLabel OS type $osType is NOT Desktop type and OS Version is $OSVersion"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_DomainJoined {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "System is not AD domain joined"
                FailureText = "System is domain joined. Unjoin from AD domain and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $script:partofDomain = (Get-WmiObject -Class win32_computersystem).partofdomain
            Write-LogFile "Computer part of domain - $script:partofDomain"
            $ImageInfo["partOfDomain"] = $script:partofDomain
            if ( $script:partofDomain -eq $true ) {
                $ErrorCode = "DomainJoined"
                Write-LogFile " $global:FailureLabel System is domain joined. Please detach the system from the domain."
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel System is not domain joined."
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_AzureDomainJoined {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "System is Azure domain joined"
                FailureText = "System is Azure domain joined. Unjoin from Azure domain and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $DomainInfo = Invoke-Expression "dsregcmd /status"
            if ($DomainInfo -match "AzureAdJoined : YES") {
                $ErrorCode = "AzureDomainJoined"
                Write-LogFile " $global:FailureLabel System is AzureAdJoined joined. Please detach the system from the domain."
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel System is not domain joined."
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_Firewall {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Windows Firewall is disabled"
                FailureText = "Windows Firewall is enabled. Turn off public firewall profile and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $location = "Registry::HKLM\SYSTEM\ControlSet001\Services\SharedAccess\Parameters\FirewallPolicy"
            $publicFirewall = Get-ItemProperty -path $location\PublicProfile -ErrorAction SilentlyContinue
            $ImageInfo["Firewall"] = @{ }
            $ImageInfo["Firewall"]["Type"] = "Public"

            if ($publicFirewall.enablefirewall -eq 1) {
                $ErrorCode = "FirewallEnabled"
                Write-LogFile " $global:FailureLabel Public Firewall Profile is turned ON"
                $ImageInfo["Firewall"]["TurnedOn"] = $TRUE
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel Public Firewall Profile is turned OFF"
                $ImageInfo["Firewall"]["TurnedOn"] = $FALSE
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            netsh advfirewall set publicprofile state off
        }
    }
}
function Test-BYOL_VMWareTools {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "VMWare tools not installed"
                FailureText = "VMWare tools are currently installed. Uninstall VMWare tools and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $path = 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\*'
            $installedApps = Get-ItemProperty $path | Select-Object DisplayName
            foreach ($app in $installedApps) {
                $ImageInfo["VMWareToolsInstalled"] = $app.displayname
                if ($app.displayname -eq "VMWare Tools") {
                    $ErrorCode = "VMWareToolsInstalled"
                    Write-LogFile " $global:FailureLabel VMware Tools is installed? - $VMWareToolsInstalled"
                    return [tuple]::Create($global:FAILED, $ErrorCode)
                }
                else {
                    Write-LogFile "=$global:SuccessLabel> VMware Tools is installed? - $VMWareToolsInstalled"
                }
            }
            return [tuple]::Create($global:PASSED)
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_DriveLessThan80GB {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Attached Disk#0 is smaller than $MaxDriveSize GB"
                FailureText = "Attached Disk#0 is larger than $MaxDriveSize GB. Make them smaller than $MaxDriveSize GB and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $drive = Get-WmiObject -Class Win32_DiskDrive | Where-Object {$_.DeviceID -like "*PHYSICALDRIVE0*"}
            $ImageInfo["Drives"] = $drive
            if (($drive.size / 1GB) -gt $global:MaxDriveSize) {
                $ErrorCode = "DiskSizeExceeded"
                Write-LogFile "$global:FailureLabel Attached drive size is greater than $global:MaxDriveSize GB"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel Attached drive size is less than $global:MaxDriveSize GB"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_GPTPartitions {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "MBR/GPT partitioned volumes"
                FailureText = "Volumes are not partitioned to be compatible with Operating System. Make sure all volumes MBR partitioned for Windows 10/7 and GPT partitioned for Windows 11 and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $script:partitions = Get-WmiObject -Class win32_diskpartition
            $ImageInfo["Partitions"] = $partitions
            $convertedPartitions = $partitions | Out-String
            $convertedPartitionTypes = $partitions.type | Out-String
            Write-LogFile "List of Partitions - "
            Write-LogFile "$convertedPartitions"
            Write-LogFile "(in respective order to above partition list)"
            Write-LogFile "$convertedPartitionTypes"
            foreach ($partition in $partitions) {
                $ErrorCode = "IncompatiblePartitioning"
                if ($partition.type.startswith("GPT") -and (-not (Validate-UEFIBootEnabled))) {
                    Write-LogFile "$global:FailureLabel There is a GPT (GUID Partition Table) Partition on non UEFI enabled OS"
                    return [tuple]::Create($global:FAILED, $ErrorCode)
                }
                elseif ($partition.type.startswith("MBR") -and (Validate-UEFIBootEnabled)) {
                    Write-LogFile "$global:FailureLabel There is a MBR (Master Boot Record) Partition on UEFI enabled OS"
                    return [tuple]::Create($global:FAILED, $ErrorCode)
                }
            }
            Write-LogFile "$global:SuccessLabel $($partitions.Name) are compatible with the OS"
            return [tuple]::Create($global:PASSED)
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_PendingReboots {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "No pending system reboot"
                FailureText = "System is pending reboot. Complete system reboot and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $RebootRequired = 'HKLM:\Software\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\RebootRequired'
            $RebootPending = 'HKLM:\Software\Microsoft\Windows\CurrentVersion\Component Based Servicing\RebootPending'
            if ((test-path $RebootRequired -ErrorAction SilentlyContinue) -or (test-path $RebootPending -ErrorAction SilentlyContinue)) {
                $ErrorCode = "PendingReboot"
                Write-LogFile " $global:FailureLabel There are pending updates. All updates need to be installed before the Image can be converted to BYOL AMI"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel No pending reboots detected."
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_AutoLogon {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "AutoLogon is disabled"
                FailureText = "AutoLogon is enabled. Disable AutoLogon in registry and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $path = 'HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion\Winlogon'
            $result = Get-ItemProperty -path $path
            if ($result.AutoAdminLogon -eq 1) {
                $ErrorCode = "AutoLogonEnabled"
                Write-LogFile " $global:FailureLabel AutoLogon turned on? - True"
                $ImageInfo["AutoLogonTurnedOn"] = $TRUE
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel AutoLogon turned on? - False"
                $ImageInfo["AutoLogonTurnedOn"] = $FALSE
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            Set-ItemProperty -Path "HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion\Winlogon" -Name AutoAdminLogon -Value "0"
        }
    }
}
function Test-BYOL_RealTimeUniversal {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "RealTimeISUniversal registry key is enabled"
                FailureText = "RealTimeIsUniversal registry key is disabled. Enable this setting and try again. `nSee: `nhttp://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/windows-set-time.html"
            }
            $global:FunctionList += $obj
        }
        true {
            $path = 'HKLM:\System\CurrentControlSet\Control\TimeZoneInformation'
            $result = Get-ItemProperty -path $path
            if ($result.RealTimeIsUniversal) {
                Write-LogFile "$global:SuccessLabel RealTimeIsUniversal reg key installed - True"
                $ImageInfo["RealTimeIsUniversal"] = $TRUE
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "RealTimeUniversalDisabled"
                Write-LogFile " $global:FailureLabel RealTimeIsUniversal reg key installed - False"
                $ImageInfo["RealTimeIsUniversal"] = $FALSE
                return [tuple]::Create($global:WARNING, $ErrorCode)
            }
        }
        fix {
            reg add "HKEY_LOCAL_MACHINE\System\CurrentControlSet\Control\TimeZoneInformation" /v RealTimeIsUniversal /d 1 /t REG_DWORD /f
        }
    }
}
function Test-BYOL_MultipleBootPartition {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Single bootable partition"
                FailureText = "System has multiple bootable partitions. Decrease the number of bootable partitions to one and try again."
            }
            $global:FunctionList += $obj
        }
        true {
            $partitions = Get-WmiObject -class win32_diskpartition
            $numberofBootable = ($partitions | Group-Object Bootable) | Where-Object { $_.Name -eq "True" } | Select-Object -ExpandProperty Count
            if ($numberofBootable -gt 1) {
                $ErrorCode = "MultipleBootPartition"
                Write-LogFile " $global:FailureLabel Number of bootable partition - $numberofBootable"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel Number of bootable partition - $numberofBootable"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_64BitOS {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "OS is 64 bit"
                FailureText = "OS is not 64 bit. Try again with a 64 bit OS image."
            }
            $global:FunctionList += $obj
        }
        true {
            $OSArchitecture = (Get-WmiObject -Class Win32_ComputerSystem).SystemType
            Write-LogFile " OS Architecture - $OSArchitecture"
            $ImageInfo["OSArchitecture"] = $OSArchitecture
            if ($OSArchitecture -match 'x64') {
                Write-LogFile "$global:SuccessLabel This OS is a 64 bit OS. "
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "64BitOS"
                Write-LogFile $global:AdditionalOSRequirementsNotMet64BitLanguageLogMessage
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_RearmCount {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Rearm count is 0"
                FailureText = "The rearm count of the image must not be 0."
            }
            $global:FunctionList += $obj
        }
        true {
            # If rearm count is greatet then 0 = passed
            if ((cscript C:\Windows\System32\slmgr.vbs /dlv | Out-String) -match "windows.+: ([1-9])") {
                Write-LogFile "$global:SuccessLabel The rearm count is sufficient for BYOL."
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "ZeroRearmCount"
                Write-LogFile " $global:FailureLabel The rearm count is not sufficient for BYOL."
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_InPlaceUpgrade {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Image is not in-place upgraded"
                FailureText = "Image is in-place upgrade. Try again with fresh image."
            }
            $global:FunctionList += $obj
        }
        true {
            $upgradecheck = Get-ChildItem -Path HKLM:\SYSTEM\Setup | Where-Object { $_.PsChildName -match "Source OS" -or $_.PsChildName -match "Upgrade" }
            if ($upgradecheck -eq $null) {
                Write-LogFile "$global:SuccessLabel OS is Not Upgraded"
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "InPlaceUpgrade"
                Write-LogFile " $global:FailureLabel OS is Upgraded"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}
function Test-BYOL_AVnotInstalled {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "Antivirus is not installed"
                FailureText = "Antivirus is installed. Uninstall Antivirus $AVname."
            }
            $global:FunctionList += $obj
        }
        true {
            # No Antivirus should be installed, exept "Windows Defender"
            $AVname = (Get-WmiObject -Namespace "root\SecurityCenter2" -Class AntiVirusProduct).displayName | Where-Object { $_ -NotLike "Windows Defender" }
            if ($AVname -eq $null) {
                Write-LogFile "$global:SuccessLabel Antivirus is not installed"
                return [tuple]::Create($global:PASSED)
            }
            else {
                $ErrorCode = "AntiVirusInstalled"
                Write-LogFile " $global:FailureLabel Antivirus is installed. Detected Antivirus name $AVname."
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}

function Test-BYOL_BootMode_UEFI {
    switch ($AMIAutomation) {
        false {
            $obj = [PsCustomObject]@{
                Name        = $($MyInvocation.MyCommand.Name)
                TextLabel   = "BYOL Image Import with UEFI Bootmode"
                FailureText = "Windows 10 image with UEFI Boot mode is not supported for Image Import"
            }
            $global:FunctionList += $obj
        }
        true {
            if ((Validate-UEFIBootEnabled) -and ($global:OSCheck -like "*Windows 10*")) {
                $ErrorCode = "UEFINotSupported"
                Write-LogFile " $global:FailureLabel Windows 10 image with UEFI Boot mode is not supported for BYOL Image Import"
                return [tuple]::Create($global:FAILED, $ErrorCode)
            }
            else {
                Write-LogFile "$global:SuccessLabel The image with UEFI Boot mode is supported for BYOL Image Import"
                return [tuple]::Create($global:PASSED)
            }
        }
        fix {
            $global:OutputBox.AppendText($global:FailureText)
        }
    }
}

function Init-BYOLCheckerRegkey {
    try {
        Get-Item HKLM:\Software\Amazon\BYOLChecker -ErrorAction Stop | Out-Null
    }
    catch {
        New-Item "HKLM:\Software\Amazon\BYOLChecker" -Force | Out-Null
    }
}
function Write-BYOLCheckerRegkey {
    Init-BYOLCheckerRegkey
    New-ItemProperty "HKLM:\Software\Amazon\BYOLChecker" -Name version -Value $ScriptVersion -Force
    New-ItemProperty "HKLM:\Software\Amazon\BYOLChecker" -Name status -Value $status -Force
    New-ItemProperty "HKLM:\Software\Amazon\BYOLChecker" -Name date -Value $fileTimestamp -Force
}
function Test-RegistryValue {
    param (
        [parameter(Mandatory = $true)]
        [ValidateNotNullOrEmpty()]$Path,
        [parameter(Mandatory = $true)]
        [ValidateNotNullOrEmpty()]$Value
    )
    try {
        Get-ItemProperty -Path $Path -ErrorAction Stop | Select-Object -ExpandProperty $Value -ErrorAction Stop | Out-Null
        return $true
    }
    catch {
        return $false
    }
}
function Get-BYOLCheckerRegkey {
    $BYOLCheckerRegkey = "HKLM:\Software\Amazon\BYOLChecker"

    if (Test-RegistryValue -Path $BYOLCheckerRegkey -Value 'Status') {
        $Resp = Get-ItemProperty $BYOLCheckerRegkey
        $Message = "SUCCESS : $($Resp.status) - BYOLChecker version $($Resp.version) was executed on $($Resp.date)"
    }
    else {
        $Message = "WARNING : Offline BYOLChecker has not validated the image"
    }
    Write-LogFile -Type Information -Message $Message
}
function Write-LogFile {
    Param ([string]$logstring)
    try {
        $currentTime = Get-Date
        Add-content $Logfile -Encoding Unicode -Value "$currentTime $logstring"
    }
    catch {
        # if the logs themselves give an error in logging. Simply, writing to host.
        Write-Host $_.Exception.Message
    }
}
function Get-RandomCharacters($length, $characters) {
    # function used for generating a random password
    $random = 1..$length | ForEach-Object { Get-Random -Maximum $characters.length }
    $private:ofs = ""
    return [String]$characters[$random]
}
function Optimize-String([string]$inputString) {
    # function used for generating a random password
    $characterArray = $inputString.ToCharArray()
    $OptimizedStringArray = $characterArray | Get-Random -Count $characterArray.Length
    $outputString = -join $OptimizedStringArray
    return $outputString
}
function New-Password {
    # function used for generating a random password
    $password = Get-RandomCharacters -length 10 -characters 'abcdefghiklmnoprstuvwxyz'
    $password += Get-RandomCharacters -length 1 -characters 'ABCDEFGHKLMNOPRSTUVWXYZ'
    $password += Get-RandomCharacters -length 1 -characters '1234567890'
    $password += Get-RandomCharacters -length 1 -characters '!"§$%&/()=?}][{@#*+'
    $password = Optimize-String $password
    return $password
}
function New-WorkSpaces_BYOL_User {
    $password = New-Password
    NET USER "Workspaces_BYOL" $password /add | Out-Null
    NET LOCALGROUP "Administrators" "Workspaces_BYOL" /add | Out-Null
    $global:OutputBox.AppendText("`nNew User WorkSpaces_BYOL has been Startd.`nPlease copy and retain the following `nWorkspaces_BYOL/$password `n`n")
}
function SaveImageInfo {
    $InfoInJson = $ImageInfo | ConvertTo-Json
    Set-Content $Infofile -Encoding Unicode -Value $InfoInJson
}
function Set-EC2Config {
    # Enable execution of user data on bootup
    $configPath = "C:\Program Files\Amazon\Ec2ConfigService\Settings\config.xml"
    if (Test-Path $configPath) {
        Write-LogFile -Type Information -Message "Configuring EC2Config Settings & Config"
        $configFile = Get-Item $configPath
        $configXml = [xml](Get-Content $configFile)
        $pluginList = $configXml.Ec2ConfigurationSettings.Plugins.Plugin
        foreach ($plugin in $pluginList) {
            switch ($plugin.Name) {
                "Ec2HandleUserData" { $plugin.state = "Enabled" }
                default { }
            }
        }
        $globalSettings = $configXml.Ec2ConfigurationSettings.GlobalSettings
        $globalSettings.SetDnsSuffixList = "false"
        $configXml.save($configPath)
        Write-LogFile -Type Information -Message "Deleting old UserData Scripts"
        if (Test-Path "C:\Program Files\Amazon\Ec2ConfigService\Scripts\UserScript.ps1") {
            Remove-Item "C:\Program Files\Amazon\Ec2ConfigService\Scripts\UserScript.ps1" -Force
        }
    }
    else {
        Write-LogFile -Type Information -Message "Cannot find EC2Config config file at $configPath. Skipping EC2Config"
    }
}
function Set-EC2LaunchConfig {
    #Setting the ComputerName parameter to true.
    Write-LogFile -Type Information -Message "Updating Launch config file"
    $launchPath = "C:\ProgramData\Amazon\EC2-Windows\Launch\Config\LaunchConfig.json"
    if (Test-Path $launchPath) {
      $ConfigFile = Get-Content -Path $launchPath -Raw | ConvertFrom-Json
      if ($ConfigFile.setComputerName -ne $false) {
        $ConfigFile.setComputerName = $false
      }
      if ($ConfigFile.setWallpaper -ne $false) {
        $ConfigFile.setWallpaper = $false
      }
      if ($ConfigFile.AddDnsSuffixList -ne $false) {
          $ConfigFile.AddDnsSuffixList = $false
      }
      ConvertTo-Json -InputObject $ConfigFile | Set-Content $launchPath
  }
  else {
    Write-LogFile -Type Information -Message "Cannot find EC2Launch config file at $launchPath. Skipping EC2Launch config"
  }
}
function Set-EC2Launchv2Config {
    #Setting the ComputerName parameter to true.
    Write-LogFile -Type Information -Message "Updating EC2Launch agent-config file"
    $launchv2Path = "C:\ProgramData\Amazon\EC2Launch\config\agent-config.yml"
    if (Test-Path $launchv2Path) {
        $ConfigFile = & "C:\Program Files\Amazon\EC2Launch\EC2Launch.exe" get-agent-config --format json | ConvertFrom-Json
        $Stages = $ConfigFile.config
        for($i = 0; $i -lt $Stages.Count; $i++){
            switch($Stages.stage[$i]){
            "preReady"
            {
                $task = $Stages[$i].tasks
                for($j = 0; $j -lt $task.Count; $j++){
                    if($task[$j].task -match "setDnsSuffix"){
                        $task= $task | Where-Object{$_.task -ne "setDnsSuffix"}
                    }
                    if($task[$j].task -match "setWallpaper"){
                        $task= $task | Where-Object{$_.task -ne "setWallpaper"}
                    }
                }
                $Stages[$i].tasks = @($task)
                break
            }
            "postReady"{
                $task = $Stages[$i].tasks
                for($j = 0; $j -lt $task.Count; $j++){
                    if($task[$j].task -match "setHostName"){
                        $task= $task | Where-Object{$_.task -ne "setHostName"}
                    }
                }
                $Stages[$i].tasks = @($task)
                break
            }
            }
        }
      $ConfigFile.config = $Stages
      $ConfigFile | ConvertTo-Json -Depth 6 | Out-File -encoding UTF8 -FilePath $launchv2Path
    }
    else {
        Write-LogFile -Type Information -Message "Cannot find EC2Launchv2 config file at $launchv2Path. Skipping EC2launchv2 config"
    }
}
function Validate-UEFIBootEnabled {
    #Validate if UEFI Boot is enabled on the Instance
    $Bootmode = bcdedit | Select-String "path.*efi"
    If ($Bootmode){
        return $true
    }
    else{
        return $false
    }
}
function Test-RunAsAdmin {
    $RuAsAdmin = ([Security.Principal.WindowsPrincipal][Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole]::Administrator)
    if (!($RuAsAdmin)) {
        Add-Type -AssemblyName System.Windows.Forms
        # show the MsgBox:
        $result = [System.Windows.Forms.MessageBox]::Show(
            'Administrator rights are required to run this script.
Open a PowerShell session that is running as Administrator, and rerun the script.
For more information, see https://docs.aws.amazon.com/workspaces/latest/adminguide/byol-windows-images.html
'
            , 'Error', 'OK', 16)
        exit
    }
}
function Set-OSFunctions {
    # Create Function List based on OS Version
    switch ([environment]::OSVersion.Version.Major) {
        6 { $ExemptFunct = $WIN7ExemptedFunctions }
        10 { $ExemptFunct = $WIN10ExemptedFunctions }
    }
    $global:ListFunctions = $ListFunctionsAll | Where-Object { $ExemptFunct -NotContains $_ }
}
function Disable-Hibernation {
    # Disabling Hibernations
    powercfg.exe /hibernate off
    Write-LogFile "Disabled hibernation"
}
function Reset-Ec2LaunchV2 {
    # Reset Ec2LaunchV2 so that the state files are removed and the user-data is executed on the next launch
    $stateFilePath = "C:\ProgramData\Amazon\EC2Launch\state"
    if(Test-Path -path $stateFilePath){
        $stateFiles = Get-ChildItem -Path $stateFilePath
        If($stateFiles){
            Write-LogFile "Found Ec2LaunchV2 files. Reset Ec2LaunchV2 agent."
            $Exitcode = (Start-Process -FilePath "C:\Program Files\Amazon\EC2Launch\EC2Launch.exe" -Args "reset -c" -Wait -PassThru).ExitCode
            if($Exitcode -eq 0){
                Write-LogFile  "Reset Ec2LaunchV2 completed sucessfully"
            }else{
                Write-LogFile "Reset Ec2LaunchV2 failed with exit code $($Exitcode)"
                exit;
            }
        }else{
            Write-LogFile "Ec2LaunchV2 state files does not exist. Ec2lauchv2 reset not required"
        }
    }else{
        Write-LogFile "Ec2LaunchV2 path '$stateFilePath' does not exist. Ec2lauchv2 reset not required"
    }
}
function Set-AutomationFunctions {
    # Create Function List for AMIAutomation workflow (exclude functions not needed)
    $global:ListFunctions = $global:ListFunctions | Where-Object { $AutomationExemptFunctions -NotContains $_ }
}
# Display Results of test in the form
function ClearOutputBox {
    $global:OutputBox.clear()
}
function Button_Click {
    #start writing results to log
    Write-LogFile (get-date)

    $mainPanel.Controls.Add($progressBarLabel)
    $mainPanel.Controls.Add($progressBar)

    clearOutputBox

    #Reset Office Key
    Remove-ItemProperty "HKLM:\Software\Amazon\BYOLChecker" -Name SkipOfficePopup -Force -ea SilentlyContinue

    #place holder text
    $global:OutputBox.AppendText("Running Tests... This could take a few minutes to complete.")

    $TestResults = New-Object System.Collections.ArrayList
    For ($i = 0; $i -lt $global:CHECK_COUNT; $i++) {
        $ResultLabels.Item($i).forecolor = "Black"
        $ResultLabels.Item($i).text = "Checking.."
        $CommandName = $ListFunctions.Get($i)
        Write-LogFile "Beginning test : ${CommandName}"
        try {
            if ( ${CommandName} -eq "checkHotFixesInstalled") {
                $global:OutputBox.AppendText("`n`Running Tests... UI may be unresponsive for this section.. Please Wait..")
            }
            $scriptResult = (Get-Item "function:$CommandName").ScriptBlock.Invoke();
            if ($CommandName -eq "checkHotFixesInstalled") {
                clearOutputBox
                $global:OutputBox.AppendText("Running Tests... This could take a few minutes to complete")
            }
        }
        catch {
            Write-LogFile $_
            if ($scriptResult -eq $null) {
                Write-LogFile " ${CommandName} failed with an exception"
                $scriptResult = $global:FAILED;
            }
        }
        Write-LogFile "Ending test : ${CommandName}"
        $TestResults.Add($scriptResult.Item1)
        $progressBar.value = (100 / $global:CHECK_COUNT) * ($i + 1)
        switch ($TestResults.Item($i)) {
            PASSED {
                $ResultLabels.Item($i).ForeColor = "Green"
                $ResultLabels.Item($i).Text = $global:PASSED
                $TroubleButtons.Item($i).Visible = $false
            }
            FAILED {
                $ResultLabels.Item($i).ForeColor = "Red"
                $ResultLabels.Item($i).Text = $global:FAILED
                $FailureText = $FailureLabelTexts.Get($i)
                $ClickFunction = {
                    $global:OutputBox.clear()
                    $global:OutputBox.AppendText($FailureText)
                }.GetNewClosure()
                $TroubleButtons.Item($i).Visible = $true
                $TroubleButtons.Item($i).Add_Click($ClickFunction)
                $FailedFlag ++
                Write-Host "Failed function: $CommandName"
            }
            WARNING {
                $ResultLabels.Item($i).ForeColor = "DarkOrange"
                $ResultLabels.Item($i).Text = $global:WARNING
                $FailureText = $FailureLabelTexts.Get($i)
                $ClickFunction = {
                    $global:OutputBox.clear()
                    $global:OutputBox.AppendText($FailureText)
                }.GetNewClosure()

                $ClickFunctionFix = {
                    $AMIAutomation = "fix"
                    Get-Date | Out-Host
                    $ListFunctions.Get($i) | Out-Host
                    $global:BYOLFunctions.Get($i).Name
                    $global:OutputBox.clear()
                    $global:OutputBox.AppendText("Fixing")
                }.GetNewClosure()

                $TroubleButtons.Item($i).Visible = $true
                $TroubleButtons.Item($i).Add_Click($ClickFunction)

                $global:failedIndexList = $failedIndexList + $i
                $FixButtons.Item($i).Add_Click( { button_Click2 $failedIndexList })
                $ButtonFix.Visible = $true
                if (-not ($CommandName -in $warningExemptedFunctions)) {
                    $WarningFlag ++
                }
            }
            Default {
                $global:OutputBox.AppendText("Unable to return results")
            }
        }
        $script:form.Refresh()
    }
    clearOutputBox
    $global:OutputBox.AppendText("Done")
    SaveImageInfo
    if ($FailedFlag -gt 0 -or $WarningFlag -gt 0) {
        $status = "fail"
    }
    else {
        $status = "success"
    }

    # EC2Launchv2 / EC2Launch/ EC2Config agent to be configured based on operating system
    if ($global:OSCheck -like "*Server 2016*" -and $global:OSCheck -like "*Server 2019*"){
        Set-EC2LaunchConfig
    }
    elseif ((Validate-UEFIBootEnabled) -or ($global:OSCheck -like "*Windows 11*")){
        Set-EC2Launchv2Config
    }
    else{
        Set-EC2Config
    }

    Write-BYOLCheckerRegkey

    If ($status -eq "success") {
        $mainPanel.Visible = $false
        # New Panel
        $Panel2.Controls.Add($script:TextBox2)
        $Panel2.Controls.Add($script:textBox3)
        $Panel2.Controls.Add($script:textBox4)
        $Panel2.Controls.Add($script:textBox5)
        $Panel2.Controls.Add($script:textBox6)
        $Panel2.Controls.Add($script:textBox7)
        $Panel2.Controls.Add($script:textBox8)
        $Panel2.Controls.Add($script:textBox9)
        $Panel2.Controls.Add($script:textBox10)
        $Panel2.Controls.Add($LinkLabel3)
        $Panel2.Controls.Add($LinkLabel4)
        $Panel2.Controls.Add($ButtonSys)
        $Panel2.Visible = $true
    }
    $mainPanel.Controls.Remove($progressBar)
}
function Button_Click2 {
    clearOutputBox
    $AMIAutomation = "fix"
    foreach ($failedIndex in $failedIndexList) {
        & $ListFunctions.Get($failedIndex) #-AMIAutomation fix
        $outputbox.appendtext("`nFixing $($ListFunctions.Get($failedIndex))")
    }
    $outputbox.appendtext("`nFixing Completed. `nRerun the Tests")
    $Button1.Text = "Rerun Tests"
    $ButtonFix.Visible = $false
}
function Button_Click3 {
    #clearOutputBox
    Run-Sysprep
    $script:SysprepErrorFileEntry = "Running Sysprep "
    $global:sysprepExe = "success"

    if ($global:sysprepExe -eq "success") {
        # New Panel
        $Panel2.Visible = $false
        $Text = "Sysprep started"
        $TextBox = New-TexBox 0 40 500 30 $Text "29, 129, 2" "Arial" 14 Regular "TopLeft"
        $Panel3.Controls.Add($TextBox)
        $Panel3.Visible = $true
        $script:t = 0
        $outputbox1.appendtext("$script:SysprepErrorFileEntry")
        do {
            $script:t++
            Get-SysprepLogs
            Start-Sleep 2
        }
        while ($script:t -lt $global:sysprepTimeRun)
    }
}
# Display Results of test to form
# Form section
function New-Form {
    Add-Type -AssemblyName System.Windows.Forms
    Add-Type -AssemblyName System.Drawing

    # Changing AMIAutomation to true to run the functions
    $AMIAutomation = $true
    $TestLabelTexts = $FunctionList.TextLabel
    $FailureLabelTexts = $FunctionList.FailureText
    $global:CHECK_COUNT = $FunctionList.Count

    # Build Form
    $Form = New-Object System.Windows.Forms.Form
    $Form.Text = "Amazon WorkSpaces Image Validation"
    $Form.Size = New-Object System.Drawing.Size(550, 900)
    $Form.MaximumSize = New-Object System.Drawing.Size(550, 1000)
    $form.FormBorderStyle = "Sizable"
    $Form.StartPosition = "WindowsDefaultLocation"
    $Form.Topmost = $False
    $Form.ShowInTaskbar = $True
    $Form.AutoSizeMode = "GrowAndShrink"
    $Form.SizeGripStyle = "auto"
    $Form.AutoScroll = $True
    $iconPath = "${PSScriptRoot}\workspacesIcon.ico"
    $icon = [system.drawing.icon]::ExtractAssociatedIcon($iconPath)
    $Form.Icon = $icon
    $script:form = $Form;
    # Main Panel
    $mainPanel = new-panel mainPanel
    $Form.Controls.Add($mainPanel)

    $Panel2 = new-panel panel2
    $Form.Controls.Add($Panel2)

    $Panel3 = new-panel Panel3
    $Form.Controls.Add($Panel3)

    # Big buttons
    # Add big buttons
    $Button1 = New-Object System.Windows.Forms.Button
    $Button1.Anchor = "top" , "left"
    $locX = $mainPanel.Location.X
    $locY = $mainPanel.Location.Y + 22
    $Button1.Location = New-Object System.Drawing.Point $locX , $locY
    $Button1.Size = New-Object System.Drawing.Size(100, 30)
    $Button1.Text = "Begin Tests"
    $Font = New-Object System.Drawing.Font("Arial", 10, [System.Drawing.FontStyle]::Bold)
    $Button1.Font = $Font
    $Button1.ForeColor = "255, 255, 255"
    $Button1.BackColor = "67, 96, 140"

    $ButtonFix = New-Object System.Windows.Forms.Button
    $ButtonFix.Anchor = "top" , "left"
    $locX = $mainPanel.Location.X + 105
    $locY = $mainPanel.Location.Y + 22
    $ButtonFix.Location = New-Object System.Drawing.Point $locX , $locY
    $ButtonFix.Size = New-Object System.Drawing.Size(130, 30)
    $ButtonFix.Text = "Fix All Warnings"
    $Font = New-Object System.Drawing.Font("Arial", 10, [System.Drawing.FontStyle]::Bold)
    $ButtonFix.Font = $Font
    $ButtonFix.BackColor = "Orange"
    $ButtonFix.Visible = $false

    $ButtonSys = New-Object System.Windows.Forms.Button
    $ButtonSys.Anchor = "top" , "left"
    $locX = $Panel2.Location.X
    $locY = $Panel2.Location.Y + 325
    $ButtonSys.Location = New-Object System.Drawing.Point $locX , $locY
    $ButtonSys.Size = New-Object System.Drawing.Size(100, 30)
    $ButtonSys.Text = "Run Sysprep"
    $Font = New-Object System.Drawing.Font("Arial", 10, [System.Drawing.FontStyle]::Bold)
    $ButtonSys.Font = $Font
    $ButtonSys.ForeColor = "255, 255, 255"
    $ButtonSys.BackColor = "67, 96, 140"
    $ButtonSys.Visible = $true

    $progressBarLabel = New-Object System.Windows.Forms.label
    $progressBarLabel.Anchor = "top", "right"
    $locX = $mainPanel.Location.X + 0.55 * $mainPanel.Size.Width
    $locY = $mainPanel.Location.Y + 5
    $progressBarLabel.Location = New-Object System.Drawing.Point $locX , $locY
    $progressBarLabel.AutoSize = $True
    $progressBar = New-Object System.Windows.Forms.ProgressBar
    $progressBar.Anchor = "top", "right"
    $progressBar.Value = 0
    $progressBar.Style = "Continuous"
    $locX = $mainPanel.Location.X + 0.5 * $mainPanel.Size.Width
    $locY = $mainPanel.Location.Y + 22
    $progressBar.Location = New-Object System.Drawing.Point $locX , $locY
    $progressBar.Size = New-Object System.Drawing.Size(170, 30)

    #Add big buttons to form
    $mainPanel.Controls.Add($Button1)
    $mainPanel.Controls.Add($ButtonFix)
    $mainPanel.Controls.Add($script:textBox1)

    #Add Button event
    $Button1.Add_Click( { Button_Click })
    $ButtonFix.Add_Click( { Button_Click2 })
    $ButtonSys.Add_Click( { Button_Click3 })

    # Output Box
    $global:OutputBox = New-Object System.Windows.Forms.RichTextBox
    $locX = $mainPanel.Location.X
    $locY = $mainPanel.Location.Y + 82
    $global:OutputBox.Location = New-Object System.Drawing.Point $locX , $locY
    $global:OutputBox.Size = New-Object System.Drawing.Size(420, 60)
    $mainPanel.Controls.Add($global:OutputBox)

    # Output Box1
    $global:OutputBox1 = New-Object System.Windows.Forms.RichTextBox
    $locX = $Panel3.Location.X
    $locY = $Panel3.Location.Y + 100
    $global:OutputBox1.Location = New-Object System.Drawing.Point $locX , $locY
    $global:OutputBox1.Size = New-Object System.Drawing.Size(500, 400)
    $Panel3.Controls.Add($global:OutputBox1)

    # Test Labels
    $TestLabels = New-Object System.Collections.ArrayList

    For ($i = 0; $i -lt $global:CHECK_COUNT; $i++) {
        $TestLabel = New-Object System.Windows.Forms.label
        $TestLabel.Anchor = "top" , "left"
        $locX = $mainPanel.Location.X
        $locY = $mainPanel.Location.Y + 130 + ($i + 1) * 22
        $TestLabel.Location = New-Object System.Drawing.Point $locX, $locY
        $TestLabel.Text = $TestLabelTexts.Get($i)
        $TestLabel.AutoSize = $true
        $TestLabels.Add($TestLabel)
        $mainPanel.Controls.Add($TestLabel)
    }
    # Results Labels
    $ResultLabels = New-Object System.Collections.ArrayList
    #Add results labels
    For ($i = 0; $i -lt $global:CHECK_COUNT; $i++) {
        $ResultLabel = New-Object System.Windows.Forms.label
        $ResultLabel.Anchor = "top", "right"
        $locX = $mainPanel.Location.X + 0.65 * $mainPanel.Size.Width
        $locY = $mainPanel.Location.Y + 130 + ($i + 1) * 22
        $ResultLabel.Location = New-Object System.Drawing.Point $locX , $locY
        $ResultLabel.Text = "To be tested"
        $ResultLabel.AutoSize = $True
        $ResultLabels.Add($ResultLabel)
        $mainPanel.Controls.Add($ResultLabel)
    }
    $TroubleButtons = New-Object System.Collections.ArrayList
    # Add info buttons
    For ($i = 0; $i -lt $global:CHECK_COUNT; $i++) {
        $TroubleButton = New-Object System.Windows.Forms.Button
        $TroubleButton.Anchor = "top" , "right"
        $locX = $mainPanel.Location.X + 0.80 * $mainPanel.Size.Width
        $locY = $mainPanel.Location.Y + 130 + ($i + 1) * 22
        $TroubleButton.Location = New-Object System.Drawing.Point $locX , $locY
        $TroubleButton.Size = New-Object System.Drawing.Size(50, 20)
        $TroubleButton.Text = "Info..."
        $TroubleButton.Visible = $false
        $TroubleButtons.Add($TroubleButton)
        $mainPanel.Controls.Add($TroubleButton)
    }
    # Add fix buttons
    $FixButtons = New-Object System.Collections.ArrayList
    For ($i = 0; $i -lt $global:CHECK_COUNT; $i++) {
        $FixButton = New-Object System.Windows.Forms.Button
        $FixButton.Anchor = "top" , "right"
        $locX = $mainPanel.Location.X + 0.89 * $mainPanel.Size.Width
        $locY = $mainPanel.Location.Y + 130 + ($i + 1) * 22
        $FixButton.Location = New-Object System.Drawing.Point $locX , $locY
        $FixButton.Size = New-Object System.Drawing.Size(50, 20)
        $FixButton.Text = "Fix..."
        $FixButton.Visible = $false
        $FixButtons.Add($FixButton)
        $mainPanel.Controls.Add($FixButton)
    }
    $lastTestLocation = [System.Windows.Forms.label]$TestLabels[-1]
    # Link Labels
    # Add Link labels
    $LinkLabel1 = New-Object System.Windows.Forms.LinkLabel
    $locX = $mainPanel.Location.X
    $locY = $lastTestLocation.Location.Y + 35
    $LinkLabel1.Location = New-Object System.Drawing.Point $locX , $locY
    $LinkLabel1.AutoSize = $True
    $LinkLabel1.LinkColor = "BLUE"
    $LinkLabel1.ActiveLinkColor = "RED"
    $LinkLabel1.Text = "AWS VM Import Prerequisite Page"
    $LinkLabel1.add_Click( { [system.Diagnostics.Process]::start("http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/VMImportPrerequisites.html") })

    $LinkLabel2 = New-Object System.Windows.Forms.LinkLabel
    $locX = $mainPanel.Location.X
    $locY = $lastTestLocation.Location.Y + 52
    $LinkLabel2.Location = New-Object System.Drawing.Point $locX , $locY
    $LinkLabel2.AutoSize = $True
    $LinkLabel2.LinkColor = "BLUE"
    $LinkLabel2.ActiveLinkColor = "RED"
    $LinkLabel2.Text = "Windows Management Framework 4.0 (For All Other Windows OS)"
    $LinkLabel2.add_Click( { [system.Diagnostics.Process]::start("https://www.microsoft.com/en-au/download/details.aspx?id=40855") })

    $Text = "Note: All tests must PASS for WorkSpaces image validation to succeed"
    $script:textBox1 = New-TexBox 10 62 500 20 $Text "black" "Arial" 9 Bold "TopLeft"

    $Text = "Validation successful"
    $script:textBox2 = New-TexBox 0 40 500 20 $Text "29, 129, 2" "Arial" 14 Regular "TopLeft"

    $Text = "Review the following locale settings:"
    $script:textBox3 = New-TexBox 0 70 500 20 $Text "black" "Arial" 12 Regular "TopLeft"

    $Text = "$([char]0x2022) System Locale :`n$([char]0x2022) User Locale:"
    $script:textBox4 = New-TexBox 0 95 140 50 $Text "black" "Arial" 12 Regular "TopLeft"

    $Text = "$global:SystemLocale`n$global:UserLocale"
    $script:textBox5 = New-TexBox 140 95 60 50 $Text "black" "Arial" 12 Regular "TopLeft"

    $Text = "Proceed to run the Sysprep"
    $script:textBox6 = New-TexBox 0 210 500 20 $Text "black" "Arial" 12 Bold "TopLeft"

    $Text = "Once Sysprep completes, your system will shut down"
    $script:textBox7 = New-TexBox 0 240 400 20 $Text "black" "Arial" 12 Regular "TopLeft"

    $Text = "Follow the remaining instructions to complete importing BYOL image"
    $script:textBox8 = New-TexBox 0 260 510 20 $Text "black" "Arial" 12 Regular "TopLeft"

    $LinkLabel4 = New-Object System.Windows.Forms.LinkLabel
    $locX = 0
    $locY = 280
    $LinkLabel4.Location = New-Object System.Drawing.Point $locX , $locY
    $LinkLabel4.AutoSize = $True
    $LinkLabel4.LinkColor = "BLUE"
    $LinkLabel4.ActiveLinkColor = "RED"
    $LinkLabel4.Text = "BYOL Administration Guide"
    $Font = New-Object System.Drawing.Font("Arial", 12, [System.Drawing.FontStyle]::Regular)
    $LinkLabel4.Font = $Font
    $LinkLabel4.TextAlign = "TopLeft"
    $LinkLabel4.add_Click( { [system.Diagnostics.Process]::start("https://docs.aws.amazon.com/workspaces/latest/adminguide/byol-windows-images.html") })

    $Text = "I want to change locale settings"
    $script:textBox9 = New-TexBox 0 400 500 20 $Text "black" "Arial" 12 Bold "TopLeft"

    $LinkLabel3 = New-Object System.Windows.Forms.LinkLabel
    $locX = 0
    $locY = 420
    $LinkLabel3.Location = New-Object System.Drawing.Point $locX , $locY
    $LinkLabel3.AutoSize = $True
    $LinkLabel3.LinkColor = "BLUE"
    $LinkLabel3.ActiveLinkColor = "RED"
    $LinkLabel3.Text = "View guide"
    $Font = New-Object System.Drawing.Font("Arial", 12, [System.Drawing.FontStyle]::Regular)
    $LinkLabel3.Font = $Font
    $LinkLabel3.add_Click( { [system.Diagnostics.Process]::start("https://msdn.microsoft.com/en-us/library/windows/hardware/dn965674(v=vs.85).aspx") })

    $LinkLabel5 = New-Object System.Windows.Forms.LinkLabel
    $LinkLabel5.AutoSize = $True
    $LinkLabel5.LinkColor = "BLUE"
    $LinkLabel5.ActiveLinkColor = "RED"
    $LinkLabel5.Text = "Sysprep Logs"
    $LinkLabel5.add_Click( { [system.Diagnostics.Process]::start("C:\Windows\System32\sysprep\Panther\setuperr.log") })

    # Add link labels to form
    $mainPanel.Controls.Add($LinkLabel1)
    $mainPanel.Controls.Add($LinkLabel2)
    $mainPanel.Controls.Add($script:TextBox1)

    # Show the Form (do this at the end of the function)
    $form.ShowDialog() | Out-Null
}
function New-Panel ($pannelName) {
    New-Variable -Name $pannelName -Force -Scope Global
    $global:pannelName = New-Object Windows.Forms.Panel
    $global:pannelName.Anchor = "top", "left"
    $global:pannelName.Size = New-Object System.Drawing.Size (500, 840)
    $locX = $global:pannelName.Location.X + 10
    $locY = $global:pannelName.Location.Y + 5
    $global:pannelName.Location = New-Object System.Drawing.Point $locX, $locY
    return $global:pannelName
}
Function New-TexBox ($x, $y, $sizeX, $sizeY, $text, $color, $font, $fontSize, $FontStyle, $textAlign) {
    # function to create Text Box for Form
    $TextBox = New-Object System.Windows.Forms.Label
    $locX = $Panel.Location.X + $x
    $locY = $Panel.Location.Y + $y
    $TextBox.Location = New-Object System.Drawing.Point $locX , $locY
    $TextBox.Size = New-Object System.Drawing.Size($sizeX, $sizeY)
    $Font = New-Object System.Drawing.Font($Font, $fontSize, [System.Drawing.FontStyle]::$FontStyle)
    $TextBox.Font = $Font
    $TextBox.TextAlign = $textAlign
    $TextBox.ForeColor = $color
    $TextBox.Text = $text
    return $TextBox
}
function Get-OOBE {
    $systemInfo = Get-WmiObject Win32_OperatingSystem
    $global:RegisteredOrganization = [string]$systemInfo.Organization
    $global:RegisteredOwner = [string]$systemInfo.RegisteredUser
    $global:SystemLocale = Get-Culture | Select -ExpandProperty name
    $global:UserLocale = Get-UICulture | Select -ExpandProperty name
}
function Set-OOBE {
    $xml = [xml](Get-Content $global:OOBEfile)
    $xmlElement = $xml.get_DocumentElement()
    $xmlSetting = (($xmlElement.settings | Where-Object { $_.Pass -eq "oobeSystem" }).Component | Where-Object { $_.Name -eq "Microsoft-Windows-International-Core" })
    $xmlSetting.InputLocale = "$global:SystemLocale"
    $xmlSetting.SystemLocale = "$global:SystemLocale"

    $xmlSetting.UILanguage = "$global:UserLocale"
    $xmlSetting.UserLocale = "$global:UserLocale"

    $xmlSetting = (($xmlElement.settings | Where-Object { $_.Pass -eq "oobeSystem" }).Component | Where-Object { $_.Name -eq "Microsoft-Windows-Shell-Setup" })
    $xmlSetting.RegisteredOrganization = $RegisteredOrganization
    $xmlSetting.RegisteredOwner = $RegisteredOwner
    $xmlSetting = (($xmlElement.settings | Where-Object { $_.Pass -eq "specialize" }).Component | Where-Object { $_.Name -eq "Microsoft-Windows-Shell-Setup" })
    $xmlSetting.RegisteredOrganization = $RegisteredOrganization

    $xml.save($global:OOBEfile)
}
function Run-Sysprep {
    Set-OOBE
    # Clear sysprep logs before running it
    $script:SysprepErrorFileLocation = "C:\Windows\system32\sysprep\Panther\setuperr.log"
    if (test-path $script:SysprepErrorFileLocation) {
        Remove-Item $script:SysprepErrorFileLocation -Force -ErrorAction SilentlyContinue
    }
    $script:SysprepSuccededTag = "C:\Windows\system32\sysprep\Sysprep_succeeded.tag"
    if (test-path $script:SysprepSuccededTag) {
        Remove-Item $script:SysprepSuccededTag -Force -ErrorAction SilentlyContinue
    }
    Write-host "Syspreping ..."
    if ((Get-Service -Name Schedule).Status -ne "Running") {
        Start-Service Schedule
    }
    # The task is ran once on demand to execute sysprep under system context to mimic IMS process and will delete on completion
    $argList = "/RL HIGHEST /Create /RU `"SYSTEM`" /TN `"sysprep`" /SC `"ONEVENT`" /F /EC Application /MO *[System/EventID=777777] /TR `"'C:\Windows\System32\Sysprep\sysprep.exe' /oobe /generalize /shutdown /quiet /unattend:$global:OOBEfile`""
    Start-process "C:\Windows\System32\schtasks.exe" -ArgumentList $argList -Wait -WindowStyle Hidden
    Start-process "C:\Windows\System32\schtasks.exe" -ArgumentList "/Run /tn `"sysprep`"" -Wait -WindowStyle Hidden
    Start-process "C:\Windows\System32\schtasks.exe" -ArgumentList "/Delete /tn `"sysprep`" /f" -WindowStyle Hidden
}
function Get-SysprepLogs {

    If ((Test-Path $script:SysprepErrorFileLocation) -and $null -ne (Get-Content -Path $script:SysprepErrorFileLocation)) {
        $Panel3.Controls.Remove($TextBox)
        $Text = "Sysprep failed"
        $TextBox11 = New-TexBox 0 40 200 30 $Text "red" "Arial" 14 Regular "TopLeft"
        $Panel3.Controls.Add($TextBox11)
        $Panel3.Visible = $true

        $global:OutputBox1.clear()
        $script:SysprepErrorFileEntry = ((Get-Content -Path $SysprepErrorFileLocation) | Select-String -Pattern "error", "fail" | Select-Object -Last 10)
        $outputbox1.Controls.Add($LinkLabel5)

        foreach ($line in $script:SysprepErrorFileEntry) {
            $text = "`n`n$line"
            $outputbox1.appendtext($text)
        }

        $script:t = $global:sysprepTimeRun
        $outputbox1.appendtext("`n`nOpening sysprep error log file in a separate window")
        Invoke-Item $SysprepErrorFileLocation
    }
    if (Test-Path $script:SysprepSuccededTag) {
        $Panel3.Controls.Remove($TextBox11)
        $Text = "Sysprep succeeded"
        $TextBox12 = New-TexBox 0 40 200 30 $Text "green" "Arial" 14 Regular "TopLeft"
        $Panel3.Controls.Add($TextBox12)
        $Panel3.Visible = $true
        $global:OutputBox1.clear()
        $outputbox1.appendtext("Sysprep succeeded. System will shutdown in 5 sec")
        shutdown.exe /s /f /t 5
    }
    else {
        $outputbox1.appendtext(" .")
    }
}

function main {
    # get all functions that will validate the system
    try {
        $ListFunctionsAll = (Get-BYOLFunctions).Name
        # AMI Automation condition is used for AMI Automation Workflow
        Set-OSFunctions
        if ($AMIAutomation) {
            . "${PSScriptRoot}\helper.ps1"
            Set-AutomationFunctions
            if (!([System.Diagnostics.EventLog]::SourceExists($scriptName))) {
                New-EventLog -LogName $LogName -Source $scriptname
            }
            else {
                Write-Debug "Event Log Source $scriptname exist"
            }

            Get-BYOLCheckerRegkey
            Init-BYOLCheckerRegkey
            New-ItemProperty -Path "HKLM:\Software\Amazon\BYOLChecker" -Name SkipOfficePopup -Value 1 -Force | Out-Null
            foreach ($byolFunc in $global:ListFunctions) {
                # Execute each Test BYOL function to validate the system
                # This will create a list of function objects with the function name and Text Labels needed to condtruct the Form

                $result = Invoke-Expression $byolFunc
                Write-LogFile -Type Information -Message "$result : $byolFunc"

                # Storing the results Success/Failed
                $resultList += $result.Item1
                if ($result.Item1 -like "FAILED" -or $result.Item1 -like "WARNING") {
                    $errorCodesList += @($result.Item2)
                }
            }

            # Check the string for any failure
            # A failure will be recorded in DDB with Ami Automation Workflow and will stop AMI automation build
            if ($resultList -match "FAILED" -or $resultList -match "WARNING") {
                $BYOLValue = [tuple]::Create("FAILED", $errorCodesList)
            }
            else {
                $BYOLValue = [tuple]::Create("success")
            }
            return $BYOLValue
        }
        # Manual interactive form AMIAutomation = $true
        else {
            $interactiveMode = $true
            Init-BYOLCheckerRegkey
            foreach ($byolFunc in $global:ListFunctions) {
                Invoke-Expression $byolFunc
            }
            # From the above Function Objects List Start the Interactive Form
            Write-Host "Begin Interactive Form Checker"
            Test-RunAsAdmin
            Disable-Hibernation
            if($global:OSCheck -like "*Windows 11*"){
                Reset-Ec2LaunchV2
            }
            Get-OOBE
            New-Form | Out-Null
        }
    }
    catch {
        Write-Host "Unexpected error: $_"
    }
}
# Execute
main

# SIG # Begin signature block
# MIInEQYJKoZIhvcNAQcCoIInAjCCJv4CAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUYFslP3a4UEZNgZt/02zCyDsV
# wkaggiC5MIIFjTCCBHWgAwIBAgIQDpsYjvnQLefv21DiCEAYWjANBgkqhkiG9w0B
# AQwFADBlMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYD
# VQQLExB3d3cuZGlnaWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVk
# IElEIFJvb3QgQ0EwHhcNMjIwODAxMDAwMDAwWhcNMzExMTA5MjM1OTU5WjBiMQsw
# CQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cu
# ZGlnaWNlcnQuY29tMSEwHwYDVQQDExhEaWdpQ2VydCBUcnVzdGVkIFJvb3QgRzQw
# ggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQC/5pBzaN675F1KPDAiMGkz
# 7MKnJS7JIT3yithZwuEppz1Yq3aaza57G4QNxDAf8xukOBbrVsaXbR2rsnnyyhHS
# 5F/WBTxSD1Ifxp4VpX6+n6lXFllVcq9ok3DCsrp1mWpzMpTREEQQLt+C8weE5nQ7
# bXHiLQwb7iDVySAdYyktzuxeTsiT+CFhmzTrBcZe7FsavOvJz82sNEBfsXpm7nfI
# SKhmV1efVFiODCu3T6cw2Vbuyntd463JT17lNecxy9qTXtyOj4DatpGYQJB5w3jH
# trHEtWoYOAMQjdjUN6QuBX2I9YI+EJFwq1WCQTLX2wRzKm6RAXwhTNS8rhsDdV14
# Ztk6MUSaM0C/CNdaSaTC5qmgZ92kJ7yhTzm1EVgX9yRcRo9k98FpiHaYdj1ZXUJ2
# h4mXaXpI8OCiEhtmmnTK3kse5w5jrubU75KSOp493ADkRSWJtppEGSt+wJS00mFt
# 6zPZxd9LBADMfRyVw4/3IbKyEbe7f/LVjHAsQWCqsWMYRJUadmJ+9oCw++hkpjPR
# iQfhvbfmQ6QYuKZ3AeEPlAwhHbJUKSWJbOUOUlFHdL4mrLZBdd56rF+NP8m800ER
# ElvlEFDrMcXKchYiCd98THU/Y+whX8QgUWtvsauGi0/C1kVfnSD8oR7FwI+isX4K
# Jpn15GkvmB0t9dmpsh3lGwIDAQABo4IBOjCCATYwDwYDVR0TAQH/BAUwAwEB/zAd
# BgNVHQ4EFgQU7NfjgtJxXWRM3y5nP+e6mK4cD08wHwYDVR0jBBgwFoAUReuir/SS
# y4IxLVGLp6chnfNtyA8wDgYDVR0PAQH/BAQDAgGGMHkGCCsGAQUFBwEBBG0wazAk
# BggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEMGCCsGAQUFBzAC
# hjdodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURS
# b290Q0EuY3J0MEUGA1UdHwQ+MDwwOqA4oDaGNGh0dHA6Ly9jcmwzLmRpZ2ljZXJ0
# LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcmwwEQYDVR0gBAowCDAGBgRV
# HSAAMA0GCSqGSIb3DQEBDAUAA4IBAQBwoL9DXFXnOF+go3QbPbYW1/e/Vwe9mqyh
# hyzshV6pGrsi+IcaaVQi7aSId229GhT0E0p6Ly23OO/0/4C5+KH38nLeJLxSA8hO
# 0Cre+i1Wz/n096wwepqLsl7Uz9FDRJtDIeuWcqFItJnLnU+nBgMTdydE1Od/6Fmo
# 8L8vC6bp8jQ87PcDx4eo0kxAGTVGamlUsLihVo7spNU96LHc/RzY9HdaXFSMb++h
# UD38dglohJ9vytsgjTVgHAIDyyCwrFigDkBjxZgiwbJZ9VVrzyerbHbObyMt9H5x
# aiNrIv8SuFQtJ37YOtnwtoeW/VvRXKwYw02fc7cBqZ9Xql4o4rmUMIIGrjCCBJag
# AwIBAgIQBzY3tyRUfNhHrP0oZipeWzANBgkqhkiG9w0BAQsFADBiMQswCQYDVQQG
# EwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNl
# cnQuY29tMSEwHwYDVQQDExhEaWdpQ2VydCBUcnVzdGVkIFJvb3QgRzQwHhcNMjIw
# MzIzMDAwMDAwWhcNMzcwMzIyMjM1OTU5WjBjMQswCQYDVQQGEwJVUzEXMBUGA1UE
# ChMORGlnaUNlcnQsIEluYy4xOzA5BgNVBAMTMkRpZ2lDZXJ0IFRydXN0ZWQgRzQg
# UlNBNDA5NiBTSEEyNTYgVGltZVN0YW1waW5nIENBMIICIjANBgkqhkiG9w0BAQEF
# AAOCAg8AMIICCgKCAgEAxoY1BkmzwT1ySVFVxyUDxPKRN6mXUaHW0oPRnkyibaCw
# zIP5WvYRoUQVQl+kiPNo+n3znIkLf50fng8zH1ATCyZzlm34V6gCff1DtITaEfFz
# sbPuK4CEiiIY3+vaPcQXf6sZKz5C3GeO6lE98NZW1OcoLevTsbV15x8GZY2UKdPZ
# 7Gnf2ZCHRgB720RBidx8ald68Dd5n12sy+iEZLRS8nZH92GDGd1ftFQLIWhuNyG7
# QKxfst5Kfc71ORJn7w6lY2zkpsUdzTYNXNXmG6jBZHRAp8ByxbpOH7G1WE15/teP
# c5OsLDnipUjW8LAxE6lXKZYnLvWHpo9OdhVVJnCYJn+gGkcgQ+NDY4B7dW4nJZCY
# OjgRs/b2nuY7W+yB3iIU2YIqx5K/oN7jPqJz+ucfWmyU8lKVEStYdEAoq3NDzt9K
# oRxrOMUp88qqlnNCaJ+2RrOdOqPVA+C/8KI8ykLcGEh/FDTP0kyr75s9/g64ZCr6
# dSgkQe1CvwWcZklSUPRR8zZJTYsg0ixXNXkrqPNFYLwjjVj33GHek/45wPmyMKVM
# 1+mYSlg+0wOI/rOP015LdhJRk8mMDDtbiiKowSYI+RQQEgN9XyO7ZONj4KbhPvbC
# dLI/Hgl27KtdRnXiYKNYCQEoAA6EVO7O6V3IXjASvUaetdN2udIOa5kM0jO0zbEC
# AwEAAaOCAV0wggFZMBIGA1UdEwEB/wQIMAYBAf8CAQAwHQYDVR0OBBYEFLoW2W1N
# hS9zKXaaL3WMaiCPnshvMB8GA1UdIwQYMBaAFOzX44LScV1kTN8uZz/nupiuHA9P
# MA4GA1UdDwEB/wQEAwIBhjATBgNVHSUEDDAKBggrBgEFBQcDCDB3BggrBgEFBQcB
# AQRrMGkwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBBBggr
# BgEFBQcwAoY1aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0VHJ1
# c3RlZFJvb3RHNC5jcnQwQwYDVR0fBDwwOjA4oDagNIYyaHR0cDovL2NybDMuZGln
# aWNlcnQuY29tL0RpZ2lDZXJ0VHJ1c3RlZFJvb3RHNC5jcmwwIAYDVR0gBBkwFzAI
# BgZngQwBBAIwCwYJYIZIAYb9bAcBMA0GCSqGSIb3DQEBCwUAA4ICAQB9WY7Ak7Zv
# mKlEIgF+ZtbYIULhsBguEE0TzzBTzr8Y+8dQXeJLKftwig2qKWn8acHPHQfpPmDI
# 2AvlXFvXbYf6hCAlNDFnzbYSlm/EUExiHQwIgqgWvalWzxVzjQEiJc6VaT9Hd/ty
# dBTX/6tPiix6q4XNQ1/tYLaqT5Fmniye4Iqs5f2MvGQmh2ySvZ180HAKfO+ovHVP
# ulr3qRCyXen/KFSJ8NWKcXZl2szwcqMj+sAngkSumScbqyQeJsG33irr9p6xeZmB
# o1aGqwpFyd/EjaDnmPv7pp1yr8THwcFqcdnGE4AJxLafzYeHJLtPo0m5d2aR8XKc
# 6UsCUqc3fpNTrDsdCEkPlM05et3/JWOZJyw9P2un8WbDQc1PtkCbISFA0LcTJM3c
# HXg65J6t5TRxktcma+Q4c6umAU+9Pzt4rUyt+8SVe+0KXzM5h0F4ejjpnOHdI/0d
# KNPH+ejxmF/7K9h+8kaddSweJywm228Vex4Ziza4k9Tm8heZWcpw8De/mADfIBZP
# J/tgZxahZrrdVcA6KYawmKAr7ZVBtzrVFZgxtGIJDwq9gdkT/r+k0fNX2bwE+oLe
# Mt8EifAAzV3C+dAjfwAL5HYCJtnwZXZCpimHCUcr5n8apIUP/JiW9lVUKx+A+sDy
# Divl1vupL0QVSucTDh3bNzgaoSv27dZ8/DCCBrAwggSYoAMCAQICEAitQLJg0pxM
# n17Nqb2TrtkwDQYJKoZIhvcNAQEMBQAwYjELMAkGA1UEBhMCVVMxFTATBgNVBAoT
# DERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8GA1UE
# AxMYRGlnaUNlcnQgVHJ1c3RlZCBSb290IEc0MB4XDTIxMDQyOTAwMDAwMFoXDTM2
# MDQyODIzNTk1OVowaTELMAkGA1UEBhMCVVMxFzAVBgNVBAoTDkRpZ2lDZXJ0LCBJ
# bmMuMUEwPwYDVQQDEzhEaWdpQ2VydCBUcnVzdGVkIEc0IENvZGUgU2lnbmluZyBS
# U0E0MDk2IFNIQTM4NCAyMDIxIENBMTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCC
# AgoCggIBANW0L0LQKK14t13VOVkbsYhC9TOM6z2Bl3DFu8SFJjCfpI5o2Fz16zQk
# B+FLT9N4Q/QX1x7a+dLVZxpSTw6hV/yImcGRzIEDPk1wJGSzjeIIfTR9TIBXEmtD
# mpnyxTsf8u/LR1oTpkyzASAl8xDTi7L7CPCK4J0JwGWn+piASTWHPVEZ6JAheEUu
# oZ8s4RjCGszF7pNJcEIyj/vG6hzzZWiRok1MghFIUmjeEL0UV13oGBNlxX+yT4Us
# SKRWhDXW+S6cqgAV0Tf+GgaUwnzI6hsy5srC9KejAw50pa85tqtgEuPo1rn3MeHc
# reQYoNjBI0dHs6EPbqOrbZgGgxu3amct0r1EGpIQgY+wOwnXx5syWsL/amBUi0nB
# k+3htFzgb+sm+YzVsvk4EObqzpH1vtP7b5NhNFy8k0UogzYqZihfsHPOiyYlBrKD
# 1Fz2FRlM7WLgXjPy6OjsCqewAyuRsjZ5vvetCB51pmXMu+NIUPN3kRr+21CiRshh
# WJj1fAIWPIMorTmG7NS3DVPQ+EfmdTCN7DCTdhSmW0tddGFNPxKRdt6/WMtyEClB
# 8NXFbSZ2aBFBE1ia3CYrAfSJTVnbeM+BSj5AR1/JgVBzhRAjIVlgimRUwcwhGug4
# GXxmHM14OEUwmU//Y09Mu6oNCFNBfFg9R7P6tuyMMgkCzGw8DFYRAgMBAAGjggFZ
# MIIBVTASBgNVHRMBAf8ECDAGAQH/AgEAMB0GA1UdDgQWBBRoN+Drtjv4XxGG+/5h
# ewiIZfROQjAfBgNVHSMEGDAWgBTs1+OC0nFdZEzfLmc/57qYrhwPTzAOBgNVHQ8B
# Af8EBAMCAYYwEwYDVR0lBAwwCgYIKwYBBQUHAwMwdwYIKwYBBQUHAQEEazBpMCQG
# CCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQQYIKwYBBQUHMAKG
# NWh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydFRydXN0ZWRSb290
# RzQuY3J0MEMGA1UdHwQ8MDowOKA2oDSGMmh0dHA6Ly9jcmwzLmRpZ2ljZXJ0LmNv
# bS9EaWdpQ2VydFRydXN0ZWRSb290RzQuY3JsMBwGA1UdIAQVMBMwBwYFZ4EMAQMw
# CAYGZ4EMAQQBMA0GCSqGSIb3DQEBDAUAA4ICAQA6I0Q9jQh27o+8OpnTVuACGqX4
# SDTzLLbmdGb3lHKxAMqvbDAnExKekESfS/2eo3wm1Te8Ol1IbZXVP0n0J7sWgUVQ
# /Zy9toXgdn43ccsi91qqkM/1k2rj6yDR1VB5iJqKisG2vaFIGH7c2IAaERkYzWGZ
# gVb2yeN258TkG19D+D6U/3Y5PZ7Umc9K3SjrXyahlVhI1Rr+1yc//ZDRdobdHLBg
# XPMNqO7giaG9OeE4Ttpuuzad++UhU1rDyulq8aI+20O4M8hPOBSSmfXdzlRt2V0C
# FB9AM3wD4pWywiF1c1LLRtjENByipUuNzW92NyyFPxrOJukYvpAHsEN/lYgggnDw
# zMrv/Sk1XB+JOFX3N4qLCaHLC+kxGv8uGVw5ceG+nKcKBtYmZ7eS5k5f3nqsSc8u
# pHSSrds8pJyGH+PBVhsrI/+PteqIe3Br5qC6/To/RabE6BaRUotBwEiES5ZNq0RA
# 443wFSjO7fEYVgcqLxDEDAhkPDOPriiMPMuPiAsNvzv0zh57ju+168u38HcT5uco
# P6wSrqUvImxB+YJcFWbMbA7KxYbD9iYzDAdLoNMHAmpqQDBISzSoUSC7rRuFCOJZ
# DW3KBVAr6kocnqX9oKcfBnTn8tZSkP2vhUgh+Vc7tJwD7YZF9LRhbr9o4iZghurI
# r6n+lB3nYxs6hlZ4TjCCBsAwggSooAMCAQICEAxNaXJLlPo8Kko9KQeAPVowDQYJ
# KoZIhvcNAQELBQAwYzELMAkGA1UEBhMCVVMxFzAVBgNVBAoTDkRpZ2lDZXJ0LCBJ
# bmMuMTswOQYDVQQDEzJEaWdpQ2VydCBUcnVzdGVkIEc0IFJTQTQwOTYgU0hBMjU2
# IFRpbWVTdGFtcGluZyBDQTAeFw0yMjA5MjEwMDAwMDBaFw0zMzExMjEyMzU5NTla
# MEYxCzAJBgNVBAYTAlVTMREwDwYDVQQKEwhEaWdpQ2VydDEkMCIGA1UEAxMbRGln
# aUNlcnQgVGltZXN0YW1wIDIwMjIgLSAyMIICIjANBgkqhkiG9w0BAQEFAAOCAg8A
# MIICCgKCAgEAz+ylJjrGqfJru43BDZrboegUhXQzGias0BxVHh42bbySVQxh9J0J
# dz0Vlggva2Sk/QaDFteRkjgcMQKW+3KxlzpVrzPsYYrppijbkGNcvYlT4DotjIdC
# riak5Lt4eLl6FuFWxsC6ZFO7KhbnUEi7iGkMiMbxvuAvfTuxylONQIMe58tySSge
# TIAehVbnhe3yYbyqOgd99qtu5Wbd4lz1L+2N1E2VhGjjgMtqedHSEJFGKes+JvK0
# jM1MuWbIu6pQOA3ljJRdGVq/9XtAbm8WqJqclUeGhXk+DF5mjBoKJL6cqtKctvdP
# bnjEKD+jHA9QBje6CNk1prUe2nhYHTno+EyREJZ+TeHdwq2lfvgtGx/sK0YYoxn2
# Off1wU9xLokDEaJLu5i/+k/kezbvBkTkVf826uV8MefzwlLE5hZ7Wn6lJXPbwGqZ
# IS1j5Vn1TS+QHye30qsU5Thmh1EIa/tTQznQZPpWz+D0CuYUbWR4u5j9lMNzIfMv
# wi4g14Gs0/EH1OG92V1LbjGUKYvmQaRllMBY5eUuKZCmt2Fk+tkgbBhRYLqmgQ8J
# JVPxvzvpqwcOagc5YhnJ1oV/E9mNec9ixezhe7nMZxMHmsF47caIyLBuMnnHC1mD
# jcbu9Sx8e47LZInxscS451NeX1XSfRkpWQNO+l3qRXMchH7XzuLUOncCAwEAAaOC
# AYswggGHMA4GA1UdDwEB/wQEAwIHgDAMBgNVHRMBAf8EAjAAMBYGA1UdJQEB/wQM
# MAoGCCsGAQUFBwMIMCAGA1UdIAQZMBcwCAYGZ4EMAQQCMAsGCWCGSAGG/WwHATAf
# BgNVHSMEGDAWgBS6FtltTYUvcyl2mi91jGogj57IbzAdBgNVHQ4EFgQUYore0GH8
# jzEU7ZcLzT0qlBTfUpwwWgYDVR0fBFMwUTBPoE2gS4ZJaHR0cDovL2NybDMuZGln
# aWNlcnQuY29tL0RpZ2lDZXJ0VHJ1c3RlZEc0UlNBNDA5NlNIQTI1NlRpbWVTdGFt
# cGluZ0NBLmNybDCBkAYIKwYBBQUHAQEEgYMwgYAwJAYIKwYBBQUHMAGGGGh0dHA6
# Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBYBggrBgEFBQcwAoZMaHR0cDovL2NhY2VydHMu
# ZGlnaWNlcnQuY29tL0RpZ2lDZXJ0VHJ1c3RlZEc0UlNBNDA5NlNIQTI1NlRpbWVT
# dGFtcGluZ0NBLmNydDANBgkqhkiG9w0BAQsFAAOCAgEAVaoqGvNG83hXNzD8deNP
# 1oUj8fz5lTmbJeb3coqYw3fUZPwV+zbCSVEseIhjVQlGOQD8adTKmyn7oz/AyQCb
# Ex2wmIncePLNfIXNU52vYuJhZqMUKkWHSphCK1D8G7WeCDAJ+uQt1wmJefkJ5ojO
# fRu4aqKbwVNgCeijuJ3XrR8cuOyYQfD2DoD75P/fnRCn6wC6X0qPGjpStOq/CUkV
# NTZZmg9U0rIbf35eCa12VIp0bcrSBWcrduv/mLImlTgZiEQU5QpZomvnIj5EIdI/
# HMCb7XxIstiSDJFPPGaUr10CU+ue4p7k0x+GAWScAMLpWnR1DT3heYi/HAGXyRkj
# gNc2Wl+WFrFjDMZGQDvOXTXUWT5Dmhiuw8nLw/ubE19qtcfg8wXDWd8nYiveQclT
# uf80EGf2JjKYe/5cQpSBlIKdrAqLxksVStOYkEVgM4DgI974A6T2RUflzrgDQkfo
# QTZxd639ouiXdE4u2h4djFrIHprVwvDGIqhPm73YHJpRxC+a9l+nJ5e6li6FV8Bg
# 53hWf2rvwpWaSxECyIKcyRoFfLpxtU56mWz06J7UWpjIn7+NuxhcQ/XQKujiYu54
# BNu90ftbCqhwfvCXhHjjCANdRyxjqCU4lwHSPzra5eX25pvcfizM/xdMTQCi2NYB
# DriL7ubgclWJLCcZYfZ3AYwwggb6MIIE4qADAgECAhAMEz440gEhtqSrZSXIDers
# MA0GCSqGSIb3DQEBCwUAMGkxCzAJBgNVBAYTAlVTMRcwFQYDVQQKEw5EaWdpQ2Vy
# dCwgSW5jLjFBMD8GA1UEAxM4RGlnaUNlcnQgVHJ1c3RlZCBHNCBDb2RlIFNpZ25p
# bmcgUlNBNDA5NiBTSEEzODQgMjAyMSBDQTEwHhcNMjIxMjE1MDAwMDAwWhcNMjMx
# MjE0MjM1OTU5WjB/MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQ
# MA4GA1UEBxMHU2VhdHRsZTEZMBcGA1UEChMQQW1hem9uLmNvbSwgSW5jLjETMBEG
# A1UECxMKd29ya3NwYWNlczEZMBcGA1UEAxMQQW1hem9uLmNvbSwgSW5jLjCCAaIw
# DQYJKoZIhvcNAQEBBQADggGPADCCAYoCggGBAONORNL+ZMF+QN2098nFxawIn6Ch
# 1xOu+02QRFHMW0g3b2YhTthsa86MlerIPrOrYasIVGmpVUTH8rs13f05tv9Vw4+g
# KUxj6jBv+ByhQT6wyPgip3dAEoOZgCDVjAQt0hiQX7hcIcfYkq2j1JhApATPAz04
# M7GWFrVHj4XYjgqnqhDlNSb+cd2eZXxNR3oia5EUF1HxqxNWBCShhXK+KchASrUK
# M/jUs5UAm5qtZTlfaThBOWFACz6NvQi1dnkSuGYfOSFdFSnxvnIJO28XnIlU2u2C
# D9iCaHBYK5vupn6r8rHinqhL9MY+/AOXvQakPoe2j+tb5NhP6hNj68I7pio/PyXQ
# al8BzY6g6Xeeod9kNPCwKtZGHhOL1bXCWu4p4LJvebUfHb7AtFbLtuQehuZ8FBe1
# pGRref2hLz7Gfjm5iDSUsCJIq9WHVt+CLvvfpHuexiimdyTCK3V3jICwIM/6vKvs
# gxKrThkjq4UTDjJ0k1ev0zqpFcvKZl2z41XPlQIDAQABo4ICBjCCAgIwHwYDVR0j
# BBgwFoAUaDfg67Y7+F8Rhvv+YXsIiGX0TkIwHQYDVR0OBBYEFCmHiWzc+SZwBKNI
# Jjk9RGWx50s/MA4GA1UdDwEB/wQEAwIHgDATBgNVHSUEDDAKBggrBgEFBQcDAzCB
# tQYDVR0fBIGtMIGqMFOgUaBPhk1odHRwOi8vY3JsMy5kaWdpY2VydC5jb20vRGln
# aUNlcnRUcnVzdGVkRzRDb2RlU2lnbmluZ1JTQTQwOTZTSEEzODQyMDIxQ0ExLmNy
# bDBToFGgT4ZNaHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0VHJ1c3Rl
# ZEc0Q29kZVNpZ25pbmdSU0E0MDk2U0hBMzg0MjAyMUNBMS5jcmwwPgYDVR0gBDcw
# NTAzBgZngQwBBAEwKTAnBggrBgEFBQcCARYbaHR0cDovL3d3dy5kaWdpY2VydC5j
# b20vQ1BTMIGUBggrBgEFBQcBAQSBhzCBhDAkBggrBgEFBQcwAYYYaHR0cDovL29j
# c3AuZGlnaWNlcnQuY29tMFwGCCsGAQUFBzAChlBodHRwOi8vY2FjZXJ0cy5kaWdp
# Y2VydC5jb20vRGlnaUNlcnRUcnVzdGVkRzRDb2RlU2lnbmluZ1JTQTQwOTZTSEEz
# ODQyMDIxQ0ExLmNydDAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4ICAQDD
# VTy/xPcE6WiK/qu5T+Rdwxea7zt+SVlEywbIcDV8K7UQusWHQPmorMkodVwyqurM
# 571Vrpq1K6kDJjCrKtJYXbM7LCrk2a1fE7SdSTjKpODqJEsG2MThtEgTUHtzJgJL
# PdumNYYR2d1Zb0tO7OJ18v2qtKiy5DWBU/SIgaEtvdMhvwJcKlM/bJmo84StC+ZC
# EF8dD0stgykQTowjsnQO60yOP99hzApqfGSrSALOKRKGGdSl7+Gf7XDzV9TJxWBe
# 5GYRPoJxWNcFe73PZwmx8fxuArMSf3gmZ0VZ9CiDRaMGngPIKowiuJYD7p5FzTD4
# x7e/Hjy0ao4a9A5tGfpmXV3zAFymZP5GTlJiL0/tVnIJY1bosV4hFFKaJcWNdM6j
# BlZGcRfvHH9K8OBjka3JtZHo6tDYw6bbykPJx5LGJjbbPuxCdj6c4CGB3naTOKc3
# aoohW03yLRy2TUmJqBzCr75D3bEiOB4aELDPnLuuSsnusXd+QqKOfHI6P06cCSs5
# KauEiBTa0+Ab4yg179CxCRQsQ+UXXXbg45ibFaEu4Gf8cAY9jD4QbHtKVaMp3/tf
# 0IhsWFiRWV07JLRKhbhSadhA0wqjZXjPAadF3nl71u9tXGowJh+8QsVdVditZUaR
# 9/UlL4x4HR8kfPXd4USwu/y41h/SznnDsryq3aODljGCBcIwggW+AgEBMH0waTEL
# MAkGA1UEBhMCVVMxFzAVBgNVBAoTDkRpZ2lDZXJ0LCBJbmMuMUEwPwYDVQQDEzhE
# aWdpQ2VydCBUcnVzdGVkIEc0IENvZGUgU2lnbmluZyBSU0E0MDk2IFNIQTM4NCAy
# MDIxIENBMQIQDBM+ONIBIbakq2UlyA3q7DAJBgUrDgMCGgUAoHgwGAYKKwYBBAGC
# NwIBDDEKMAigAoAAoQKAADAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgor
# BgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUO3Kf9FbF
# AKSGuVKMs0XaOFZzeUwwDQYJKoZIhvcNAQEBBQAEggGARpEagkfTA25ytF+dzjO4
# PkouF+jnFVK8nBpUGlrwfGcl0OcfydnITtO7SZ6ZEaEQgVcyS7kTc1HDaZkzxgHI
# i2zb5H1i1iV7OXd4upJOiP4AarG1VVyYw32M4nlP8WRM/NDQmefhvSa3Eve4/e15
# Kg8h3WvFOa6WO2oIqCsdn3QACkfWR8dA96RuE9mmhNF/XAlEbaGNWRuH/pET1PTP
# om6fTyHzxJlEzHBhbsdQ8kyPlEf+VblU+8GnPHa1CoJT0anVSvzuRnyfOdSfdhic
# cmJL+VfbURmempd97qfbIhUFew2hZwTvLljxfJIuwGHk6REFKeppotgHMbFwKHQ9
# kKsTLd+SNIjCFxak6kuhOzcUCaqdXxMyysx23Pyus1WuxvJdq2kn9Mqlbl/0++Tz
# 0srqvMe5lWKdEFtB285XPW2II/YxMPITNiWns2d3oFY5egRiBZ3OKCuR13MaXRhx
# 31yYbg+VGSmVJeyUGZs07z5WHJ1ldbfbEoYdMS3TEI50oYIDIDCCAxwGCSqGSIb3
# DQEJBjGCAw0wggMJAgEBMHcwYzELMAkGA1UEBhMCVVMxFzAVBgNVBAoTDkRpZ2lD
# ZXJ0LCBJbmMuMTswOQYDVQQDEzJEaWdpQ2VydCBUcnVzdGVkIEc0IFJTQTQwOTYg
# U0hBMjU2IFRpbWVTdGFtcGluZyBDQQIQDE1pckuU+jwqSj0pB4A9WjANBglghkgB
# ZQMEAgEFAKBpMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkF
# MQ8XDTIzMDcyNTE4MDEwOFowLwYJKoZIhvcNAQkEMSIEIK6SIiFVP4iweiu9Exuu
# c9ALryNkkWF1NRIf4hZX6Cg+MA0GCSqGSIb3DQEBAQUABIICAC932fvAznRAXz+0
# A9QNbBiFBJpewYEqNAQLsDIJot/5Tc6A/xdW9EwAXC6dXmEwQGVFx91keGBZAuB9
# jsQYr9yiMV7fYr/kqbnkHOB+zJ7BaAttjz+SSBx2T3PU4co63HtYRzSuPIh8aVPF
# hwewVRd/UUSdgC3FgTQHfMwxjQbpXWHHTW8uSIjfIGtZHzhi8a+GGR0blHheVgn9
# VKsH7QvGgaZRcHxF8L2cJ3tn8q55ja2cI2PiN2jlTKrN5RM+WiJYhbuJG0ZK0PCm
# Q734nv6aeAVaUWw3du8r81mwewqroXbYZzZDtVQH/mqjO1QqKte6BUoXFToBzUic
# waXuUEj19C9UOWHhLxGSnu0ABXd+Z/00iww25iS+dxVuYZrH4SInGzdqxWYzuyP9
# xwYrtFty76eDE3KuYpItoHSfVLrhHlBCp/YOgVUQKuCZj4/U7HIwTpkKpoEIJ3iw
# opYXkrUY3LA8u/29sIVz7C9M1ZBCSU8suuaq5Lvye72K7CzBSdE/kVlwZOpX1FKv
# IfKRCqZ4b+z4YetzghXGiMTgdJN/U+zdsf7xL/GkFMx7XQY/EUWHN7KWMtzr3Xb5
# VQSlbgVOvpfYH3hYxGfDY0qDjj/ZP5Gy3eNNwajcN5zQ3Z2GQjEZnv/7fTDIvGfL
# v6ahagxzYf5ZUZw40NE3YzoiJmL/
# SIG # End signature block
